"""
The win32 module contains functions that are specific to the Windows operating system.

Functions:
    add_to_path:
        Add a program path to the system PATH environment variable on Windows systems.

    IS_ADMIN:
        A boolean value that indicates whether the current user has administrative privileges.
"""
import ctypes
import os
import winreg
from warnings import warn

from inspyre_toolbox.path_man import prepare_path
from inspyre_toolbox.sys_man.operating_system import MOD_LOGGER as PARENT_LOGGER
from inspyre_toolbox.sys_man.operating_system.common.path import separate_path_str as separate_path
from inspyre_toolbox.sys_man.operating_system.win32.system import run_as_admin
from inspyre_toolbox.sys_man.operating_system.win32.users import is_admin
from inspyre_toolbox.sys_man.operating_system.win32.windows_registry import RegistryManager

IS_ADMIN = is_admin()


MOD_LOGGER = PARENT_LOGGER.get_child('win32')
MOD_LOGGER.debug('win32 module loaded')

from inspyre_toolbox.sys_man.operating_system.win32.environment import GUI_ACCESS


def update_user_path_variable(new_path: str):
    """
    The update_user_path_variable function updates the user's PATH variable with the new program path.

    Note:
        This function is for Windows systems only. It will not work on *nix systems.

        This function will not notify the environment of changes. *You must call notify_the_environment_of_changes
        after calling this function to notify the environment of changes*.

    Parameters:
        new_path (str):
            The full path to the directory you'd like to add to your PATH environment variable.

    Returns:
        None
    """
    new_path = prepare_path(new_path)
    with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as root:  # Get the current user
        with winreg.OpenKey(root, "Environment", 0, winreg.KEY_ALL_ACCESS) as key:  # Go to the environment key
            existing_path_value = winreg.EnumValue(key, 3)[1]  # Grab the current path value

            # Takes the current path value and appends the new program path
            new_path_value = existing_path_value + new_path + ";"

            # Update path variable with the new path value
            winreg.SetValueEx(key, "PATH", 0, winreg.REG_EXPAND_SZ, new_path_value)


def notify_the_environment_of_changes():
    """
    The notify_the_environment_of_changes function notifies the environment of changes to the PATH variable.

    Returns:
        None
    """
    HWND_BROADCAST = 0xFFFF
    WM_SETTINGCHANGE = 0x1A
    SMTO_ABORTIFHUNG = 0x0002
    result = ctypes.c_long()
    SendMessageTimeoutW = ctypes.windll.user32.SendMessageTimeoutW
    SendMessageTimeoutW(HWND_BROADCAST, WM_SETTINGCHANGE, 0, u"Environment", SMTO_ABORTIFHUNG, 5000,
                        ctypes.byref(result))


def add_to_path(new_path: str, do_not_notify: bool = False, profile_file=None):
    """
    The add_to_path function adds the program path to the system's PATH variable on Windows systems.
    This allows for easy access to programs from any terminal window.

    Args:
        new_path (str):
            The full path to the directory you'd like to add to your system's PATH
            environment variable.

    Returns:
        None
    """
    if profile_file:
        warn("The profile_file parameter is not used on Windows systems. Ignoring parameter.")

    update_user_path_variable(new_path)

    if not do_not_notify:
        notify_the_environment_of_changes()
        return print(f"Added {new_path} to path on Windows, changes have been notified to the environment")

    print(f"Added {new_path} to path on Windows, please restart shell for changes to take effect")


def remove_from_path(directory: str):
    """
    The remove_from_path function removes a directory from the PATH environment variable.

    Args:
        directory (str):
            The directory to remove from the PATH environment variable.

    Returns:
        None
    """
    directory = prepare_path(directory)
    path_directories = list_path_directories()

    if directory in path_directories:
        path_directories.remove(directory)
        new_path = os.pathsep.join(path_directories)

        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as root:  # Get the current user PLUGIN_REGISTRY
            with winreg.OpenKey(root, "Environment", 0, winreg.KEY_ALL_ACCESS) as key:  # Go to the environment key
                winreg.SetValueEx(key, "PATH", 0, winreg.REG_EXPAND_SZ, new_path)

        notify_the_environment_of_changes()
        return print(f"Removed {directory} from path on Windows, changes have been notified to the environment")

    return print(f"Directory {directory} not found in PATH variable")


def list_path_directories():
    reg_man = RegistryManager('Environment')
    return separate_path(reg_man.get_value('Path'))


MOD_LOGGER.debug(f'{IS_ADMIN=}')


del is_admin

__all__ = [
        'IS_ADMIN',
    'add_to_path',
    'remove_from_path',
    'list_path_directories',
    'update_user_path_variable',
    'run_as_admin'
        ]



from prompt_toolkit import ANSI, HTML
from prompt_toolkit.shortcuts.dialogs import yes_no_dialog

from inspyre_toolbox.console_kit.prompts import MOD_LOGGER as PARENT_LOGGER
from inspyre_toolbox.log_engine import Loggable
from inspyre_toolbox.syntactic_sweets.classes.decorators import validate_type
from inspyre_toolbox.syntactic_sweets.locks import flag_lock

MOD_LOGGER = PARENT_LOGGER.get_child('dialogs')


class ConfirmationPrompt(Loggable):
    DIALOG_TITLE_STUB = 'Please Confirm '
    DIALOG_TEXT_STUB = 'Are you sure you want to '
    DIALOG_TEXT_SUFFIX = '?'
    DIALOG_BASE = yes_no_dialog

    def __init__(self, title: str, text: str, no_title_stub=False, no_text_stub=False, no_text_suffix=False):

        super().__init__(MOD_LOGGER)

        self.__burnt = False
        self._burning = None
        self.__title = None
        self.__text = None

        self.__no_title_stub = None
        self.__no_text_stub = None

        self.no_title_stub = no_title_stub
        self.no_text_stub = no_text_stub

        title = title.title()

        if not self.no_title_stub:
            self.title = f'{self.DIALOG_TITLE_STUB}{title}'
        else:
            self.title = title

        text = f'{text.capitalize()}{"" if no_text_suffix else self.DIALOG_TEXT_SUFFIX}'

        self.text = text if self.no_text_stub else f'{self.DIALOG_TEXT_STUB}{text}'

    @property
    def answer(self) -> bool:
        return self.__answer if self.burnt else None

    @answer.setter
    @validate_type(bool)
    def answer(self, new: bool):
        if self.burning:
            self.__answer = new
            return

        if not self.burnt:
            raise AttributeError('This prompt has not been presented yet.')
        else:
            raise AttributeError('This property can only be set while the prompt is running and has not been run yet.')

    @property
    def burning(self):
        return self._burning

    @burning.setter
    @validate_type(bool)
    def burning(self, new: bool):
        self._burning = new

    @property
    def burnt(self) -> bool:
        return self.__burnt

    @burnt.setter
    @validate_type(bool)
    def burnt(self, new: bool):
        self.check_burnt()
        self.__burnt = new

    @property
    def no_text_stub(self) -> bool:
        return self.__no_text_stub

    @no_text_stub.setter
    @validate_type(bool)
    def no_text_stub(self, new: bool):
        self.check_burnt()
        self.__no_text_stub = new

    @property
    def no_title_stub(self) -> bool:
        return self.__no_title_stub

    @no_title_stub.setter
    @validate_type(bool)
    def no_title_stub(self, new: bool):
        self.check_burnt()
        self.__no_title_stub = new

    @property
    def title(self) -> str:
        return self.__title

    @title.setter
    @validate_type(str)
    def title(self, new: str):
        self.check_burnt()
        self.__title = new

    @property
    def text(self) -> str:
        self.check_burnt()
        return self.__text

    @text.setter
    @validate_type(str)
    def text(self, new: str):
        self.check_burnt()
        self.__text = new

    def check_burnt(self):
        if self.burnt and not self.burning:
            raise AttributeError('This prompt has already been burnt.')

    def run(self) -> bool:
        log = self.create_logger()
        log.debug(f'Running confirmation prompt: {self.title} with text {self.text}')
        if self.burning:
            raise RuntimeError('This prompt is already running.')

        # Log the type of self.text
        log.debug(f'Type of self.text: {type(self.text)}')

        # Ensure self.text is of the correct type
        if not isinstance(self.text, (str, HTML, ANSI)):
            raise ValueError(f'Invalid text type: {type(self.text)}. Expected str, HTML, ANSI, or FormattedText.')

        self.burnt = True
        with flag_lock(self, 'burning'):
            self.answer = yes_no_dialog(title=self.title, text=self.text).run()

            return self.answer



import contextlib
import re
from typing import Literal

from inspyre_toolbox.ver_man.classes.pypi import PyPiVersionInfo, TestPyPiVersionInfo

RELEASE_MAP = {
    'dev': 'Development Build',
    'alpha': 'Alpha Build',
    'beta': 'Beta Build',
    'rc': 'Release Candidate Build',
    'final': 'Final Release Build'
}



class VersionParser:

    RELEASE_MAP = RELEASE_MAP

    def __init__(self, version_str):
        self.version_str = version_str
        self.version_info = self.parse_version()

    @staticmethod
    def get_release(release_type: Literal[tuple[RELEASE_MAP.keys()]]) -> str:
        """
        Gets the release name from the release type.
        """
        if not isinstance(release_type, str):
            raise TypeError(f'Release type must be a string, not {type(release_type)}')

        release_type = release_type.lower()

        if release_type not in VersionParser.RELEASE_MAP:
            raise ValueError(f'Invalid release type: {release_type}. Must be one of '
                             f'{tuple(VersionParser.RELEASE_MAP.keys())}')

        return VersionParser.RELEASE_MAP.get(release_type)

    def parse_version(self):
        """
        Parses the version string and returns the version information.

        Returns:
            dict:
                A dictionary containing the version information.
        """
        parts = self.version_str.split('-')
        version_numbers = parts[0].split('.')
        major, minor, patch = map(int, version_numbers)

        release = 'final'
        release_num = 0
        release_type = 'final'  # Default release type

        if len(parts) > 1:
            release_info = parts[1]
            release_info_parts = re.split(r'[\.\+]', release_info)
            release_type = release_info_parts[0]
            release = self.RELEASE_MAP.get(release_type, 'final')

            if len(release_info_parts) > 1 and release_info_parts[1].isdigit():
                release_num = int(release_info_parts[1])

        return {
            'major': major,
            'minor': minor,
            'patch': patch,
            'release': release,
            'release_abbr': release_type,
            'release_num': release_num
        }

    def to_full_version_string(self):
        """
        Returns a formatted full-version string based on the version information.

        Returns:
            str:
                The formatted full-version string.
        """
        major, minor, patch = self.version_info['major'], self.version_info['minor'], self.version_info['patch']
        release, release_num = self.version_info['release'], self.version_info['release_num']

        version_str = f"v{major}.{minor}.{patch}"
        version_str += (
            f" {release}"
            if release not in ('final', 'Final Release Build')
            else ''
        )
        if release_num > 0:
            version_str += f" {release_num}"

        return version_str

    def print_version(self, skip_rich=False):
        """
        Prints the version information using the specified method.

        Parameters:
            skip_rich (bool):
                If True, skips the rich print method and prints the version information using the default print
                method. Default is False.

        Returns:
            None
        """
        if not skip_rich:
            with contextlib.suppress(ImportError):
                self.__rich__()
                return
        self._print_version()

    def _print_version(self):
        version_info = self.version_info
        print(version_info)

    def __str__(self):
        return self.version_str

    def __repr__(self):
        return f"VersionParser('{self.version_str}\n')"

    def __rich__(self):
        from rich.table import Table
        from rich import box
        table = Table(box=box.SIMPLE)
        table.add_column('Major', justify='right', style='cyan')
        table.add_column('Minor', justify='right', style='magenta')
        table.add_column('Patch', justify='right', style='green')
        table.add_column('Release', justify='right', style='yellow')
        table.add_column('Release Num', justify='right', style='pink')
        table.add_row(str(self.version_info['major']), str(self.version_info['minor']), str(self.version_info['patch']),
                      self.version_info['release'], str(self.version_info['release_num']))
        return table

    @property
    def full_version_string(self):
        """
        Property that returns the formatted full version-string based on the version information.

        Returns:
            str:
                The formatted full-version string.
        """
        return self.to_full_version_string()

    @property
    def major(self):
        """
        Property that returns the major version number.

        Returns:
            int:
                The major version number
        """
        return self.version_info['major']

    @property
    def minor(self):
        """
        Property that returns the minor version number.

        Returns:
            int:
                The minor version number
        """
        return self.version_info['minor']

    @property
    def patch(self):
        """
        Property that returns the patch version number.

        Returns:
            int:
                The patch version number
        """
        return self.version_info['patch']

    @property
    def release(self):
        """
        Property that returns the release type.

        Returns:
            str:
                The release type
        """
        return self.version_info['release']

    @property
    def release_num(self):
        """
        Property that returns the release number.

        Returns:
            int:
                The release number (0 if not specified)
        """
        return self.version_info['release_num']

    # Arithmetic operations
    def __add__(self, other):
        if isinstance(other, VersionParser):
            return VersionParser(f"{self.major + other.major}.{self.minor + other.minor}.{self.patch + other.patch}")
        elif isinstance(other, (int, float)):
            return VersionParser(f"{self.major}.{self.minor}.{self.patch + int(other)}")
        else:
            raise TypeError(
                f"Unsupported operand type(s) for +: 'VersionParser' and '{type(other)}'"
            )

    def __sub__(self, other):
        if isinstance(other, VersionParser):
            return VersionParser(f"{self.major - other.major}.{self.minor - other.minor}.{self.patch - other.patch}")
        elif isinstance(other, (int, float)):
            return VersionParser(f"{self.major}.{self.minor}.{self.patch - int(other)}")
        else:
            raise TypeError(
                f"Unsupported operand type(s) for -: 'VersionParser' and '{type(other)}'"
            )

    # Comparison operations
    def __eq__(self, other):
        if isinstance(other, VersionParser):
            return self.version_info == other.version_info
        return False

    def __lt__(self, other):
        if isinstance(other, VersionParser):
            return (
                (self.major, self.minor, self.patch, self.version_info['release_abbr'], self.release_num)
                < (other.major, other.minor, other.patch, other.version_info['release_abbr'], other.release_num)
            )
        return NotImplemented

    def __le__(self, other):
        return self == other or self < other

    def __gt__(self, other):
        if isinstance(other, VersionParser):
            return (
                (self.major, self.minor, self.patch, self.version_info['release_abbr'], self.release_num)
                > (other.major, other.minor, other.patch, other.version_info['release_abbr'], other.release_num)
            )
        return NotImplemented

    def __ge__(self, other):
        return self == other or self > other


__all__ = [
    'VersionParser',
    'PyPiVersionInfo',
    'TestPyPiVersionInfo',
    'RELEASE_MAP'
]



import sys

import requests
from packaging import version as pkg_version
from rich.console import Console
from rich.table import Table

from inspyre_toolbox.ver_man.classes.pypi.errors import PyPiPackageNotFoundError
from inspyre_toolbox.ver_man.classes.pypi.helpers import load_pypi_version_info

CONSOLE = Console()

BASE_URL = 'https://pypi.org/pypi/'
TEST_PYPI_BASE_URL = 'https://test.pypi.org/pypi/'


class PyPiVersionInfo:
    """
    A class to represent the version information for a package from PyPi.

    Attributes:
        package_name (str):
            The name of the package on PyPi.
        latest_stable (packaging.version.Version):
            The latest stable version of the package on PyPi.
        latest_pre_release (packaging.version.Version):
            The latest pre-release version of the package.
    Since:
        v1.6.0
    """

    def __init__(self, package_name, include_pre_release_for_update_check=False):
        """
        Initialize the PyPiVersionInfo object.

        Args:
            package_name:
            include_pre_release_for_update_check:
        """
        self.package_name = package_name
        if not hasattr(self, '_url'):
            self._url = f'{BASE_URL}{self.package_name}/json'
        self.__installed = self.get_installed_version()
        self.__checked_for_update = False
        self.__newer_available_version = None
        self.__latest_stable = None
        self.__latest_pre_release = None
        self.__all_versions = None
        self.include_pre_release_for_update_check = include_pre_release_for_update_check
        self.__query_versions()

    @property
    def all_versions(self):
        if self.__all_versions is None:
            self.__query_versions()
        return sorted([pkg_version.parse(v) for v in self.__all_versions])

    @property
    def checked_for_update(self):
        return self.__checked_for_update

    def get_installed_version(self):
        """
        Gets the installed version of the package.
        This method should be implemented to get the installed version of the package.
        For example, it can use importlib.metadata or pkg_resources to find the installed version.
        """
        try:
            import importlib.metadata

            return importlib.metadata.version(self.package_name)
        except importlib.metadata.PackageNotFoundError:
            return None

    @property
    def installed(self):
        """
        Gets the installed version of the package.
        """
        if self.__installed is None:
            return None
        return pkg_version.parse(self.__installed)

    @property
    def installed_newer_than_latest(self):
        return self.installed > self.latest

    @property
    def latest(self):
        return (
                self.all_versions[-1]
                if self.include_pre_release_for_update_check
                else self.latest_stable
        )

    @property
    def latest_stable(self):
        """
        Gets the latest stable version of the package on PyPi.
        """
        if self.__latest_stable is None:
            self.__query_versions()
        return pkg_version.parse(self.__latest_stable)

    @property
    def latest_pre_release(self):
        """
        Gets the latest pre-release version of the package.
        """
        if self.__all_versions is None:
            self.__query_versions()
        pre_release_versions = [v for v in self.all_versions if v.is_prerelease]
        return pre_release_versions[-1] if pre_release_versions else None

    @property
    def newer_available_version(self):
        if self.__newer_available_version is None:
            self.check_for_update()

        return self.__newer_available_version

    def __query_versions(self):
        """
        Queries the versions from PyPi.
        """
        try:
            response = requests.get(self.url)
            response.raise_for_status()
            data = response.json()

            self.__all_versions = list(data['releases'].keys())
            self.__latest_stable = data['info']['version']
        except requests.RequestException as e:
            raise PyPiPackageNotFoundError(
                message='Package not found on PyPi.',
                skip_print=self.__class__.__name__ == 'TestPyPiVersionInfo',
            ) from e

    @property
    def update_available(self):
        """
        Checks if an update is available.
        """
        if self.__newer_available_version is None and self.installed:
            self.check_for_update()
        elif not self.installed:
            return False
        return self.__newer_available_version is not None
    
    @property
    def url(self):
        return self._url

    def check_for_update(self, include_pre_releases=False):
        latest_version = self.latest_stable

        if include_pre_releases or self.include_pre_release_for_update_check:
            latest_version = max(latest_version, self.latest_pre_release)

        if latest_version > self.installed:
            self.__newer_available_version = latest_version

        self.__checked_for_update = True

        return latest_version > self.installed

    def update(self):
        """
        Checks for updates to the package.

        Returns:
            None

        """
        local_newer_statement = 'Local version is newer than latest version. This is likely a development build.'
        print = CONSOLE.print

        if not self.installed_newer_than_latest:
            local_newer_statement = ''

        try:
            if self.update_available:
                print(f'\n\n[bold green]Update Available![/bold green] New version: '
                      f'[bold cyan]{self.newer_available_version}[/bold cyan]')
            else:
                print(f'\n\n[bold green]No update available.[/bold green] Current version: '
                      f'[bold cyan]{self.installed}[/bold cyan] {local_newer_statement}')

        except Exception as e:
            print(f'An error occurred during the update check: {str(e)}')

    def print_version_info(self):
        """
        Print version information in a formatted table.

        This function creates a table using the `Table` class and populates it with version information about the current Python environment. It then prints the table to the console.

        Parameters:
            self: The current instance of the class.

        Returns:
            None
        """

        # Create a table

        table = Table(show_header=False, show_lines=True, expand=True, border_style='bright_blue',
                      row_styles=['none', 'dim'])

        # Add columns
        table.add_column('Property', style='cyan', width=20)
        table.add_column('Value', justify='center')

        # Add rows
        table.add_row('Package Name', self.package_name)

        installed_version = str(self.installed) if self.installed else "[bold red]Not installed[/bold red]"

        table.add_row('Installed Version', installed_version)
        table.add_row('Latest Stable Version', str(self.latest_stable))
        table.add_row('Latest Pre-release Version', str(self.latest_pre_release))

        if self.update_available:
            table.add_row('Update Available', '[bold green]Yes[/bold green]')
            table.add_row('Latest Version', str(self.newer_available_version))

        table.add_row('Python Executable Path', sys.executable)
        table.add_row('Python Version', sys.version)

        CONSOLE.print(table)


class TestPyPiVersionInfo(PyPiVersionInfo):
    """
    A class to query information from test.pypi.org.
    """
    def __init__(self, package_name, include_pre_release_for_update_check=False):
        self.package_name = package_name
        self._url = f'{TEST_PYPI_BASE_URL}{self.package_name}/json'

        super().__init__(package_name, include_pre_release_for_update_check)


__all__ = [
    'PyPiVersionInfo',
    'TestPyPiVersionInfo',
    'load_pypi_version_info'
]



"""
This module provides the version of the package.
"""

from pathlib import Path

from inspyre_toolbox.ver_man.classes import PyPiVersionInfo, VersionParser as Version
from inspyre_toolbox.ver_man.helpers import read_version_file

VERSION_FILE_NAME = 'VERSION'


def get_version_file_path(version_file_name: str = VERSION_FILE_NAME) -> Path:
    """
    Gets the version-file path.

    Parameters:
        version_file_name (str, optional):
            The name of the version file. Defaults to :str:'VERSION

    Returns:
        Path:
            The version-file path.

    Since:
        v1.6.0

    Example Usage:
        >>> from inspyre_toolbox.common.about.version import get_version_file_path
        >>> file_path = get_version_file_path()
        >>> print(file_path)
        Path('path/to/version/file')
    """
    return Path(__file__).parent / version_file_name


VERSION_FILE_PATH = get_version_file_path()

VERSION = Version(read_version_file(VERSION_FILE_PATH))

VERSION_NUMBER = VERSION.parse_version()

# Clean up the namespace
del get_version_file_path
del Version
del VERSION_FILE_PATH
del read_version_file


PYPI_VERSION_INFO = PyPiVersionInfo('inspyre-toolbox')


__all__ = [
    'PYPI_VERSION_INFO',
    'VERSION',
]



from inspyre_toolbox.common.errors import InspyreToolboxError


class ISTVersionToolError(InspyreToolboxError):
    """
    Base class for all errors in the ist-version-tool program.
    """
    pass


class NonExistentCommandError(InspyreToolboxError):
    """
    Raised when an invalid command is provided to the IST-Version-Tool program.
    """
    default_message = ' is not a valid command!'

    def __init__(self, command, additional_message=None, **kwargs):
        message = f'{command}{self.default_message}'
        self.__additional_message = None
        super().__init__(message=message, **kwargs)

        if additional_message:
            self.additional_message = additional_message

    @property
    def additional_message(self):
        return self.__additional_message

    @additional_message.setter
    def additional_message(self, new):
        if not isinstance(new, str):
            raise ValueError('additional_message must be a string.')
        self.__additional_message = new.strip()



import winreg

KEY_PATHS = {
        "Environment":           r"Environment",
        "Inspyre-Toolbox":       r"Software\Inspyre-Softworks\Inspyre-Toolbox",
        r"Inspyre-Toolbox\Path": r"Software\Inspyre-Softworks\Inspyre-Toolbox\Path",
        }


class RegistryManager:
    """
    A class to manage registry keys and values for a specific path in the Windows Registry.
    """

    KEY_PATH = KEY_PATHS['Inspyre-Toolbox']

    def __init__(self, key_path: str = None):
        """
        Initialize the RegistryManager with a specific registry key path.

        Parameters:
            key_path (str, optional):
                The path to the registry key. Defaults to the class-level KEY_PATH.
        """
        self.key_path = key_path or self.KEY_PATH

    @property
    def as_dict(self):
        """
        Get the registry key as a dictionary.

        Returns:
            dict:
                The registry key as a dictionary.

                The dictionary is structured as follows:
                - Keys are the names of the registry values.
                  - The names have '_' instead of '\\'.
                - Values are the corresponding values from the registry.
                - Subkeys are included as nested dictionaries, with their names joined by underscores to the base key path.

        Example Usage:
            reg_manager = RegistryManager()
            registry_dict = reg_manager.as_dict
            print(registry_dict)

        """
        _dict = {value[0]: value[1] for value in self.list_values()}
        for key in self.list_keys():
            _dict['_'.join(key.key_path.split('\\')[2:])] = key.as_dict
        return _dict

    @property
    def key_exists(self) -> bool:
        """
        Check if the key exists in the registry.

        Returns:
            bool: Whether the key exists.
        """
        return self.check_key_exists()

    @property
    def number_of_subkeys(self) -> int:
        """
        Get the number of subkeys in the registry key.

        Returns:
            int:
                The number of subkeys.
        """
        return self.get_number_of_subkeys()

    @property
    def num_subkeys(self) -> int:
        """
        Alias for the number_of_subkeys property.

        Returns:
            int:
                The number of subkeys.

        """
        return self.number_of_subkeys

    @property
    def subkeys(self) -> list:
        """
        List the keys in the registry key.

        Returns:
            list: The keys in the registry key.
        """
        return self.list_keys()

    @property
    def full_key_path(self) -> str:
        """
        Get the full key path.

        Returns:
            str: The full key path.
        """
        return f'HKEY_CURRENT_USER\\{self.key_path}'

    def check_key_exists(self) -> bool:
        """
        Check if the key exists in the registry.

        Returns:
            bool: Whether the key exists.
        """
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, self.key_path, 0, winreg.KEY_READ):
                return True
        except FileNotFoundError:
            return False

    def create(self):
        """
        Create the key in the registry.
        """
        try:
            with winreg.CreateKey(winreg.HKEY_CURRENT_USER, self.key_path) as key:
                pass
        except OSError as e:
            if e.errno == 5:
                raise PermissionError("Access denied. Run as administrator.")
            else:
                raise e from e

    def get_number_of_subkeys(self):
        """
        Get the number of subkeys in the registry key.

        Returns:
            int:
                The number of subkeys.
        """
        return len(self.list_keys())

    def has_value(self, value_name: str) -> bool:
        """
        Check if a value exists in the registry.

        Parameters:
            value_name (str):
                The name of the value to check.

        Returns:
            bool: Whether the value exists.
        """
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, self.key_path, 0, winreg.KEY_READ) as key:
                winreg.QueryValueEx(key, value_name)
                return True
        except FileNotFoundError:
            return False
        except OSError as e:
            if e.errno == 5:
                raise PermissionError("Access denied. Run as administrator.") from e
            else:
                raise e from e

    def list_keys(self):
        """
        List the keys in the registry key.

        Returns:
            list: The keys in the registry key.
        """
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, self.key_path, 0, winreg.KEY_READ) as key:
                i = 0
                keys = []
                while True:
                    try:
                        name = winreg.EnumKey(key, i)
                        keys.append(RegistryManager(f'{self.key_path}\\{name}'))
                        i += 1
                    except OSError:
                        break
                return keys
        except FileNotFoundError:
            return None
        except OSError as e:
            if e.errno == 5:
                raise PermissionError("Access denied. Run as administrator.")
            else:
                raise e from e

    def list_key_paths(self):
        for key in self.list_keys():
            print(key.key_path)

    def list_values(self):
        """
        List the values in the registry key.

        Returns:
            list: The values in the registry key.
        """
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, self.key_path, 0, winreg.KEY_READ) as key:
                i = 0
                values = []
                while True:
                    try:
                        name, value, type = winreg.EnumValue(key, i)
                        values.append((name, value, type))
                        i += 1
                    except OSError:
                        break
                return values
        except FileNotFoundError:
            return None
        except OSError as e:
            if e.errno == 5:
                raise PermissionError("Access denied. Run as administrator.")
            else:
                raise e from e

    def set_value(self, value_name: str, value: str):
        """
        Set a value in the registry.

        Parameters:
            value_name (str):
                The name of the value to set.

            value (str):
                The value to set.
        """
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, self.key_path, 0, winreg.KEY_WRITE) as key:
                winreg.SetValueEx(key, value_name, 0, winreg.REG_SZ, value)
        except OSError as e:
            if e.errno == 5:
                raise PermissionError("Access denied. Run as administrator.")
            else:
                raise e from e

    def get_value(self, value_name: str) -> str:
        """
        Get a value from the registry.

        Parameters:
            value_name (str):
                The name of the value to get.

        Returns:
            str: The value.
        """
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, self.key_path, 0, winreg.KEY_READ) as key:
                return winreg.QueryValueEx(key, value_name)[0]
        except FileNotFoundError:
            raise KeyError(f"Value '{value_name}' not found.")
        except OSError as e:
            if e.errno == 5:
                raise PermissionError("Access denied. Run as administrator.")
            else:
                raise e from e

    def delete_value(self, value_name: str):
        """
        Delete a value from the registry.

        Parameters:
            value_name (str):
                The name of the value to delete.
        """
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, self.key_path, 0, winreg.KEY_WRITE) as key:
                winreg.DeleteValue(key, value_name)
        except FileNotFoundError:
            raise KeyError(f"Value '{value_name}' not found.")
        except OSError as e:
            if e.errno == 5:
                raise PermissionError("Access denied. Run as administrator.")
            else:
                raise e from e

    def delete_key(self, skip_confirm: bool = False):
        """
        Delete the key from the registry.
        """
        try:
            winreg.DeleteKey(winreg.HKEY_CURRENT_USER, self.key_path)
        except FileNotFoundError:
            raise KeyError(f"Registry key '{self.key_path}' not found.")
        except OSError as e:
            if e.errno == 5:
                raise PermissionError("Access denied. Run as administrator.")
            else:
                raise e from e

    def __repr__(self):
        return f"<RegistryManager(key_path='{self.key_path}')> @ {hex(id(self))} with {self.number_of_subkeys} subkeys"


def delete_key(key_path: str):
    """
    Delete a key from the registry.

    Parameters:
        key_path (str):
            The path to the key to delete.
    """
    try:
        winreg.DeleteKey(winreg.HKEY_CURRENT_USER, key_path)
    except FileNotFoundError:
        raise KeyError(f"Registry key '{key_path}' not found.")
    except OSError as e:
        if e.errno == 5:
            raise PermissionError("Access denied. Run as administrator.")
        else:
            raise e from e


def has_key(key_path: str):
    """
    Check if a key exists in the registry.

    Parameters:
        key_path (str):
            The path to the key to check.

    Returns:
        bool: Whether the key exists.
    """
    try:
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, key_path, 0, winreg.KEY_READ):
            return True
    except FileNotFoundError:
        return False
    except OSError as e:
        if e.errno == 5:
            raise PermissionError("Access denied. Run as administrator.")
        else:
            raise e from e


def set_registry_key(root, key, subkey, value):
    """
    The set_registry_key function sets a registry key.

    Args:
        root (HKEY):
            The root key to set.
        key (str):
            The key to set.
        subkey (str):
            The subkey to set.
        value (str):
            The value to set.

    Returns:
        None
    """
    with winreg.ConnectRegistry(None, root) as root_key:
        with winreg.OpenKey(root_key, key, 0, winreg.KEY_WRITE) as key:
            winreg.SetValue(key, subkey, winreg.REG_SZ, value)


def set_path_variable(value):
    set_registry_key(winreg.HKEY_CURRENT_USER, "Environment", "Path", value)


def find_registry_key(root, key, subkey):
    """
    The find_registry_key function finds a registry key.

    Args:
        root (HKEY):
            The root key to search in.
        key (str):
            The key to search for.
        subkey (str):
            The subkey to search for.

    Returns:
        The key if found, None if not found.
    """
    try:
        with winreg.ConnectRegistry(None, root) as root_key:
            with winreg.OpenKey(root_key, key, 0, winreg.KEY_READ) as key:
                try:
                    with winreg.OpenKey(key, subkey, 0, winreg.KEY_READ) as subkey:
                        return subkey
                except FileNotFoundError:
                    return None
    except FileNotFoundError:
        return None


def list_registry_keys(root, subkey):
    try:
        with winreg.OpenKey(root, subkey) as key:
            i = 0
            keys = []
            while True:
                try:
                    name, value, type = winreg.EnumValue(key, i)
                    keys.append((name, value, type))
                    i += 1
                except OSError:
                    break
            return keys
    except FileNotFoundError:
        return None


def get_registry_dict(root, subkey):
    return RegistryManager(f'{root}\\{subkey}').as_dict


def get_directories_from_registry_path():
    from inspyre_toolbox.sys_man.helpers.path import separate_path

    reg_man = RegistryManager(KEY_PATHS['Environment'])
    return separate_path(path) if (path := reg_man.as_dict['Path']) else None



import json
from pathlib import Path

from inspyre_toolbox.common.about.author import
from inspyre_toolbox.syntactic_sweets.classes.decorators import validate_type

CONFIG_SYSTEMS = {
        'path'
        }


def get_system_config_schema_path(system: str) -> Path:
    """
    Get the path to the schema for the specified system.

    Parameters:
        system (str):
            The system to get the schema for.

    Returns:
        Path:
            The path to the schema file.

    Raises:
        ValueError:
            Raised when the specified system is not supported.
    """
    if system not in CONFIG_SYSTEMS:
        raise ValueError(f"Unsupported system: '{system}'")

    return Path(__file__).parent.joinpath(f'{system}.json')


class Config:
    """
    A class to manage configuration settings from a JSON file.

    The `Config` class reads configuration data from a JSON file and dynamically creates
    classes based on the configuration parameters. Each property includes a getter and setter
    with type validation.

    Attributes:
        json_path (str):
            Path to the JSON file containing configuration data.

    Methods:
        __init__(json_path):
            Initializes the :class:`Config` instance by loading configuration from the specified JSON file.

        _load_from_json(json_path):
            Loads configuration data from the JSON file and creates classes dynamically.

        _create_property(name, description, type_):
            Creates a property with the given name, description, and type.
    """

    def __init__(self, json_path):
        """
        Initializes the `Config` instance by loading configuration from the specified JSON file.

        Parameters:
            json_path (str): Path to the JSON file containing configuration data.

        Returns:
            None
        """
        self._load_from_json(json_path)

    def _load_from_json(self, json_path):
        """
        Loads configuration data from the JSON file and creates classes dynamically.

        Parameters:
            json_path (str): Path to the JSON file containing configuration data.

        Returns:
            None
        """
        with open(json_path, 'r') as file:
            data = json.load(file)
        for key, value in data.items():
            setattr(self, f"_{key}", value['default'])
            self._create_property(key, value['description'], value['type'])

    def _create_property(self, name, description, type_):
        """
        Creates a property with the given name, description, and type.

        Parameters:
            name (str): The name of the property.
            description (str): A description of the property.
            type_ (str): The type of the property, one of 'string', 'bool', 'int', 'float', 'list', or 'dict'.

        Returns:
            None
        """
        type_map = {
                "string": str,
                "bool":   bool,
                "int":    int,
                "float":  float,
                "list":   list,
                "dict":   dict
                }
        type_cls = type_map.get(type_, str)

        def getter(self):
            return getattr(self, f"_{name}")

        def setter(self, value):
            setattr(self, f"_{name}", value)

        getter.__doc__ = f"""
        Returns the {name} attribute.

        Returns:
            {type_cls.__name__}:
                {description}
        """
        setter.__doc__ = f"""
        Sets the {name} attribute.

        Parameters:
            new ({type_cls.__name__}):
                The value to set the {name} attribute to.

        Returns:
            None
        """
        setter = validate_type(type_cls)(setter)

        prop = property(getter, setter)
        setattr(self.__class__, name, prop)


def load_sample_config(config_system: str) -> dict:
    """
    Load the sample configuration file.

    Parameters:
        config_system (str):
            The configuration system to use.

    Returns:
        dict:
            The sample configuration.
    """
    if config_system.lower() not in CONFIG_SYSTEMS:
        raise ValueError(f"Invalid config system: '{config_system}'")

    sample_config_file = Path(__file__).parent.joinpath(f'{config_system}.json')

    if not sample_config_file.exists():
        raise FileNotFoundError(f"Sample configuration file not found: '{sample_config_file}'")

    with open(sample_config_file, 'r') as file:
        sample_config = json.load(file)

    return sample_config



"""
This script provides a command-line interface to convert between different units of digital information storage.
It uses the inspyre_toolbox library to perform the conversion and to humanize the numerical output.
The user can specify the amount and units for conversion, and the script will output the result in the desired unit.
"""


# Our imports
from inspyre_toolbox.conversions.bytes import ByteConverter
from inspyre_toolbox.humanize import Numerical
from inspyre_toolbox.cli.ist_bytes_converter.arguments import Arguments
from inspyre_toolbox.cli.ist_bytes_converter.helpers import clean_unit_str, convert_if_whole_number


# Define a function to handle the main application operations.
def main(cli_args=None):
    """
    Main function that handles the conversion between different units of digital information storage.

    Parameters:
        cli_args (argparse.Namespace, optional):
            Parsed command-line arguments. If `None`, the function will parse the arguments itself.
    
    Returns:
        None
    """
    # If `cli_args` is `None` we should create an instance 
    # of the parser.
    args = Arguments().parsed if cli_args is None else cli_args  # Parse command-line arguments

    # Gather the arguments from the command-line
    from_unit = clean_unit_str(args.from_unit)  # Clean and standardize the 'from' unit
    to_unit = clean_unit_str(args.to_unit)  # Clean and standardize the 'to' unit

    # Grab our given amount and pass it to be converted to an integer if
    # it is a whole number.
    amount = convert_if_whole_number(args.amount)

    # Create a converter object.
    converter = ByteConverter(amount, from_unit)  # Initialize the ByteConverter with the specified amount and 'from' unit
   
    # Use it to convert from our starting unit to the target unit.
    result = converter.convert(to_unit)  # Perform the conversion to the 'to' unit

    # Grab our result and pass it to be converted to an integer if it is a whole number.
    result = convert_if_whole_number(result)
    
    # Format our starting string properly by;
    #   - Passing it through the `Numerical` class,
    #   - Grabbing the resulting string and break it into a list delimited by a space,
    #   - Grab the last item in the list that results from the above operation.
    from_str = Numerical(amount, from_unit, store_as_float=False).count_noun()  # Humanize the 'from' amount and unit
    
    # Format out target string properly by;
    #   - Passing it through the `Numerical` class,
    #   - Grabbing the resulting string and break it into a list delimited by a space,
    #   - Grab the last item in the list that results from the above operation.
    to_str = Numerical(result, to_unit, store_as_float=False).count_noun()  # Humanize the 'to' amount and unit

    # Output the conversion details and result
    print(f'Converting {from_str} to {to_str.split(" ")[-1]}')
    print(f'Result: {to_str}')



from inspyre_toolbox.cli.ist_version_tool.arguments import Arguments
from inspyre_toolbox.cli.ist_version_tool.errors import NonExistentCommandError
from inspyre_toolbox.console_kit import clear_console
from inspyre_toolbox.ver_man.classes.pypi import load_pypi_version_info

ARG_PARSER = Arguments()

PYPI_VER_INFO = load_pypi_version_info('inspyre-toolbox')


def handle_print_version_info():
    """
    Handle the print version info command.

    Returns:
        None
    """
    handle_version_command()


def handle_update_command():
    PYPI_VER_INFO.update()


def handle_version_command():
    """
    Handle the version command.

    Returns:
        None
    """
    clear_console()
    PYPI_VER_INFO.print_version_info()


def handle_command(args):
    if not args.command:
        return handle_print_version_info()

    command = args.command.lower()

    if command == 'version':
        handle_version_command()

    elif command == 'update':
        handle_update_command()
    else:
        raise NonExistentCommandError(command)


def main():
    args = ARG_PARSER.parse_args()

    handle_command(args)


if __name__ == '__main__':
    main()



from functools import wraps
from pathlib import Path
from typing import Union

def validate_type(*allowed_types, preferred_type=None, allowed_values=None, case_sensitive=True, conversion_funcs=None):
    """
    A decorator for validating the type and optionally the value of a value passed to a class
    property setter, with an option to convert to a preferred type if specified and to enforce
    value restrictions.

    Parameters:
        *allowed_types:
            Variable length list of allowed types for the property value.

        preferred_type (type, optional):
            The preferred type to which values should be converted if possible. (If None, no conversion is attempted)

        allowed_values (iterable, optional):
            An iterable of values that are allowed. If None, all values of the correct type are allowed.

        case_sensitive (bool, optional):
            Specifies whether string comparisons should be case-sensitive. Defaults to True. Ignored for non-string
            types.

        conversion_funcs (dict, optional):
            A dictionary mapping types to conversion functions. These functions should convert from the
            specified type to the preferred type.

    Returns:
        A decorator function for the property setter.

    Raises:
        TypeError:
            If the incoming value does not match one of the allowed types, cannot be converted to the preferred type,
            or is not in the list of allowed values.

        ValueError:
            If the value is of the correct type but not in the allowed values list.

    Example:
        >>> class MyClass:
        ...     @property
        ...     def my_property(self):
        ...         return self._my_property
        ...
        ...     @my_property.setter
        ...     @validate_type(str, allowed_values=["hello", "world"], case_sensitive=False)
        ...     def my_property(self, value):
        ...         self._my_property = value
        ...
        >>> obj = MyClass()
        >>> obj.my_property = "HELLO"  # This is accepted and set
        >>> obj.my_property = "hello"  # This is also accepted and set
        >>> obj.my_property = "Hi"  # This raises ValueError
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, value):
            if not isinstance(value, tuple(allowed_types)):
                allowed = ', '.join([t.__name__ for t in allowed_types])
                raise TypeError(f"Value must be of type {allowed}, got type {type(value).__name__}")

            if preferred_type and not isinstance(value, preferred_type):
                if conversion_funcs and type(value) in conversion_funcs:
                    try:
                        value = conversion_funcs[type(value)](value)
                    except Exception as e:
                        raise TypeError(f"Conversion failed: {e}")
                else:
                    try:
                        value = preferred_type(value)
                    except Exception as e:
                        raise TypeError(f"Could not convert value to preferred type {preferred_type.__name__}: {e}")

            if allowed_values is not None:
                if isinstance(value, str) and not case_sensitive:
                    validation_passed = any(val.lower() == value.lower() for val in allowed_values)
                else:
                    validation_passed = value in allowed_values

                if not validation_passed:
                    allowed_vals = ', '.join(map(str, allowed_values))
                    raise ValueError(f"Value must be one of {allowed_vals}, got {value}")

            return func(self, value)

        return wrapper

    return decorator


"""


Author: 
    Inspyre Softworks

Project:
    Inspyre-Toolbox

File: 
    inspyre_toolbox/ver_man/classes/pypi/helpers.py
 

Description:
"""
from inspyre_toolbox.ver_man.classes.pypi.errors import PyPiPackageNotFoundError

DOUBLE_FAIL_MSG_1 = 'Unable to find '
DOUBLE_FAIL_MSG_2 = ' on either PyPi or TestPyPi. Please check the package name and try again.'


def load_pypi_version_info(package_name, include_pre_releases_in_update_check=False, skip_test_pypi_on_fail=False):
    """
    Loads the PyPi version info for a package.

    Parameters:
        package_name (str):
            The name of the package.

        include_pre_releases_in_update_check (bool, optional):
            If True, includes pre-releases in the update check. Defaults to False.

        skip_test_pypi_on_fail (bool, optional):
            If True, skips the test PyPi on failure. Defaults to False.

    Returns:
        PyPiVersionInfo:
            The PyPi version info.

    Since:
        v1.6.0

    Example Usage:
        >>> from inspyre_toolbox.ver_man.classes.pypi.helpers import load_pypi_version_info
        >>> pypi_version_info = load_pypi_version_info('inspyre-toolbox')
        >>> print(pypi_version_info)
        PyPiVersionInfo('inspyre-toolbox')

        >>> pypi_version_info = load_pypi_version_info('some-dev-package')
        >>> print(pypi_version_info)
        TestPyPiVersionInfo('some-dev-package')
    """
    pypi = None
    from inspyre_toolbox.ver_man.classes.pypi import PyPiVersionInfo, TestPyPiVersionInfo

    final_err_msg = f'{DOUBLE_FAIL_MSG_1}"{package_name}"{DOUBLE_FAIL_MSG_2}'

    try:
        pypi = PyPiVersionInfo(package_name, include_pre_release_for_update_check=include_pre_releases_in_update_check)
    except PyPiPackageNotFoundError as e:
        print('PyPi package not found. Trying TestPyPi...')
        if not skip_test_pypi_on_fail:
            try:
                pypi = TestPyPiVersionInfo(package_name,
                                           include_pre_release_for_update_check=include_pre_releases_in_update_check)
            except PyPiPackageNotFoundError as e:
                raise PyPiPackageNotFoundError(message=final_err_msg) from e

    return pypi



# ==============================================================================
#  Copyright (c) Inspyre Softworks 2022.                                       =
#                                                                              =
#  Author:                 T. Blackstone                                       =
#  Author Email:    <t.blackstone@inspyre.tech>                                =
#  Created:              2/10/22, 9:42 PM                                      =
# ==============================================================================

import os
import random
import sys
from time import sleep

from inspyre_toolbox.log_engine import ROOT_LOGGER as PARENT_LOGGER

MOD_LOGGER = PARENT_LOGGER.get_child('console_kit')


def __get_delay(fast=False):
    possible = list(range(45, 100)) if fast else list(range(100, 301))
    return random.choice(possible) / 1000


def clear_console():
    """

    Clear the current terminal screen.

    Returns
    -------
    None.

    """
    cmd = "cls" if os.name in ("nt", "dos") else "clear"

    os.system(cmd)


def animate_typing(
        message: str,
        interval: float = None,
        skip_pre_newline: bool = False,
        skip_post_newline: bool = False,
        clear_screen: bool = False,
        fast_typer: bool = False,
        override_upper_limit: bool = False,
):
    """

    Animate typing in the terminal/console.

    Args:
        message (str):
            The message you'd like to print through a typing animation.

        interval (Optional(int|float)):
            The time (in number of seconds as an integer) that you'd like to have pass between characters being printed
            to the line. (Defaults to random intervals to more realistically simulate typing.

        skip_pre_newline:
            Don't print a newline before commencing the animation.

        skip_post_newline (Optional(bool)):
            Don't print a newline after we're finished with the typing animation. (Optional)

        clear_screen:

         fast_typer:
        override_upper_limit:

    Returns:
        None

    """
    # If we were instructed to clear the screen, do it.
    if clear_screen:
        clear_console()

    # If we weren't instructed to skip a leading newline, print the newline.
    if not skip_pre_newline:
        print("\n")

    for char in message:
        if interval is None:
            time_between = __get_delay(fast_typer)
        else:
            time_between = interval

            time_between = min(time_between, 0.301)
            if time_between <= 0:
                time_between = __get_delay()

        sleep(time_between)
        sys.stdout.write(char)

    # If we weren't instructed to skip the post-animation newline, print the newline.
    if not skip_post_newline:
        print('\n')



import os
from pathlib import Path
from typing import List, Optional, TypeVar, Union
from warnings import warn

from inspyre_toolbox.syntactic_sweets.classes.decorators import validate_type
from inspyre_toolbox.syntactic_sweets.classes.decorators.freeze import freeze_property

PathLike = TypeVar("PathLike", str, bytes, Path, None)


def __normalize_file_types(file_types: Optional[Union[str, List[str]]]) -> List[str]:
    """Normalize file types to a list of strings."""
    if file_types is None:
        return ['*']

    return [file_types] if isinstance(file_types, str) else file_types


def __normalize_ignore_dirs(ignore_dirs: Optional[List[str]], ignore_case: bool) -> set:
    """Normalize and prepare directory names to ignore."""
    if ignore_case and ignore_dirs:
        return {dir_name.lower() for dir_name in ignore_dirs}
    return set(ignore_dirs or [])


def __filter_dirs(dir_names: List[str], ignore_dirs: set, ignore_case: bool) -> List[str]:
    """Filter out directories that should be ignored."""
    if ignore_case:
        return [d for d in dir_names if d.lower() not in ignore_dirs]
    return [d for d in dir_names if d not in ignore_dirs]


def __filter_files(dir_path: Path, file_names: List[str], file_types: List[str], get_file_object) -> List[Path]:
    """Filter and gather files based on file types."""
    return [
            get_file_object(dir_path / file_name)
            for file_name in file_names
            if any(file_name.endswith(f".{ftype}") for ftype in file_types) or '*' in file_types
            ]



class ISTB_Path:

    def __init__(self, path: Union[str, Path], auto_prepare=False, **kwargs):
        self.__auto_prepare = None
        self.__path = None
        self.__prepared = False

        self.path = path
        self.auto_prepare = auto_prepare

        if self.auto_prepare:
            self.prepare(**kwargs)

    @property
    def auto_prepare(self):
        return self.__auto_prepare

    @auto_prepare.setter
    @freeze_property
    @validate_type(bool)
    def auto_prepare(self, new):
        self.__auto_prepare = new

    @property
    def exists(self):
        return self.path.exists()

    @property
    def parent(self):
        return self.path.parent

    @property
    def path(self):
        return self.__path

    @path.setter
    @freeze_property
    def path(self, new):
        self.__path = new

    @property
    def prepared(self):
        return self.__prepared

    @prepared.setter
    @freeze_property
    @validate_type(bool)
    def prepared(self, new):
        self.__prepared = new

    def prepare(self, **kwargs):
        if not self.prepared:
            self.__path = prepare_path(self.path, **kwargs)
            self.prepared = True

        return self.path

    def __str__(self):
        return str(self.path)


def create_directory(
        directory: Union[str, bytes, os.PathLike],
        do_not_prepare: bool = False,
        do_not_expand: bool = False,
        do_not_resolve: bool = False,
        do_not_convert: bool = False,
        do_not_provision: bool = False,
        do_not_check_if_exists: bool = False,
        ):
    """
        Create a directory.

        Parameters:
            directory (Union[str, Path]):
                The directory to create.

            do_not_prepare (bool):
                A flag indicating whether to prepare the path.

            do_not_expand (bool):
                A flag indicating whether to expand the path.

            do_not_resolve (bool):
                A flag indicating whether to resolve the path.

            do_not_convert (bool):
                A flag indicating whether to convert the path to a string.

            do_not_provision (bool):
                A flag indicating whether to provision the path.

            do_not_check_if_exists (bool):
                A flag indicating whether to check if the path exists.
        """
    if not do_not_prepare:
        directory = prepare_path(
                directory,
                do_not_convert=do_not_convert,
                do_not_expand=do_not_expand,
                do_not_resolve=do_not_resolve,
                do_not_provision=do_not_provision,
                do_not_check_if_exists=do_not_check_if_exists
                )

    if not isinstance(directory, Path):
        raise ValueError(f"Invalid directory: {directory}!")

    if not do_not_check_if_exists and directory.exists():
        return

    directory.mkdir(parents=True, exist_ok=True)


def check_path(
        path: Union[str, Path],
        do_not_expand: bool = False,
        do_not_resolve: bool = False,
        do_not_convert: bool = False,
        do_not_provision: bool = False,
        ) -> bool:
    """
    Check if a path is valid.

    This function checks if a path is valid. If the path is a string, it will be converted to a Path object. If the path
    is not expanded, it will be expanded. If the path is not resolved, it will be resolved. If the path is not converted
    to a string, it will be converted to a string. If the path is not provisioned, it will be provisioned.

    Note:
        Provisioning a path involves converting it to a Path object, expanding it, and resolving it.

    Parameters:
        path (Union[str, Path]):
            The path to check.

        do_not_expand (bool):
            A flag indicating whether to expand the path.

        do_not_resolve (bool):
            A flag indicating whether to resolve the path.

        do_not_convert (bool):
            A flag indicating whether to convert the path to a string.

        do_not_provision (bool):
            A flag indicating whether to provision the path.

    Returns:
        bool:
            A flag indicating whether the path is valid.
    """
    if not do_not_provision:
        path = provision_path(
                path,
                do_not_convert=do_not_convert,
                do_not_expand=do_not_expand,
                do_not_resolve=do_not_resolve
                )

    return path.exists() if isinstance(path, Path) else False


def check_directory(
        path: Union[str, Path],
        **kwargs
        ) -> bool:
    """
    Check if a directory is valid.

    This function checks if a directory is valid. If the path is a string, it will be converted to a Path object. If the
    path is not expanded, it will be expanded. If the path is not resolved, it will be resolved. If the path is not
    converted to a string, it will be converted to a string. If the path is not provisioned, it will be provisioned.

    Note:
        Provisioning a path involves converting it to a Path object, expanding it, and resolving it.

    Parameters:
        path (Union[str, Path]):
            The directory to check.

        **kwargs:
            Additional keyword arguments.

    Returns:
        bool:
            A flag indicating whether the directory is valid.
    """
    return path.is_dir() if check_path(path, **kwargs) else False


def check_file(
        path: Union[str, Path],
        **kwargs
        ) -> bool:
    """
    Check if a file is valid.

    Parameters:

        path (Union[str, Path]):
            The file to check.

        **kwargs:
            Additional keyword arguments.

    Returns:
        bool:
            A flag indicating whether the file is valid.
    """
    return path.is_file() if check_path(path, **kwargs) else False


def prepare_path(
        path: Union[str, Path],
        do_not_expand: bool = False,
        do_not_resolve: bool = False,
        do_not_convert: bool = False,
        do_not_provision: bool = False,
        do_not_check_if_exists: bool = False,
        do_not_create: bool = False,
        return_as_string: bool = False
        ):
    """
    Prepare a path.

    Parameters:
        path (Union[str, Path]):
            The path to prepare.

        do_not_create (bool):
            A flag indicating whether to create the path.

        do_not_expand (bool):
            A flag indicating whether to expand the path.

        do_not_resolve (bool):
            A flag indicating whether to resolve the path.

        do_not_convert (bool):
            A flag indicating whether to convert the path to a string.

        do_not_provision (bool):
            A flag indicating whether to provision the path.

        do_not_check_if_exists (bool):
            A flag indicating whether to check if the path exists.

        return_as_string:
            A flag indicating whether to return the path as a string.

    Returns:
        Union[str, Path]:
            The prepared path.

    Raises:
        ValueError: If the path is invalid.

    Example:
        >>> prepare_path('~/path/to/file', return_as_string=True)
        '/home/taylor/path/to/file'

        >>> prepare_path('~/path/to/file', return_as_string=False)
        PosixPath('/home/taylor/path/to/file')

    """
    if not do_not_provision:
        path = provision_path(
                path,
                do_not_convert=do_not_convert,
                do_not_expand=do_not_expand,
                do_not_resolve=do_not_resolve
                )

    if path.suffix:
        do_not_create = True
        warn(f"File path detected: {path}", UserWarning)

    if not do_not_check_if_exists:

        exists = check_path(path)

        if not exists and do_not_create:
            raise ValueError(f"Invalid path: {path}")
        elif not exists:
            create_directory(path, do_not_prepare=True, do_not_provision=True, do_not_check_if_exists=True)

    return str(path) if return_as_string else path


def provision_path(
        path: Union[str, Path],
        do_not_expand: bool = False,
        do_not_resolve: bool = False,
        do_not_convert: bool = False,
        ) -> Path:
    """
    Provision a path.

    Parameters:
        path (str):
            The path to provision.

        do_not_convert (bool):
            A flag indicating whether to convert the path to a string.

        do_not_expand (bool):
            A flag indicating whether to expand the path.

        do_not_resolve (bool):
            A flag indicating whether to resolve the path.

    Returns:
        Path:
            The provisioned path.
    """
    if not isinstance(path, Path):
        if isinstance(path, str) and not do_not_convert:
            path = Path(path)
        else:
            raise ValueError(f"Invalid path: {path}!")

    if not do_not_expand:
        path = path.expanduser()

    if not do_not_resolve:
        path = path.resolve()

    return path


def provision_paths(path_list: list) -> list:
    """
    Provision a list of paths.

    This function takes a list of paths and provisions them.

    Note:
        Provisioning a path involves converting it to a Path object, expanding it, and resolving it.

    Parameters:
        path_list (list):
            The list of paths to provision.

    Returns:
        list:
            The provisioned list of paths.
    """
    return [provision_path(path) for path in path_list]


def gather_files_in_dir(
        directory: Union[str, Path],
        recursive: bool = False,
        file_types: Optional[Union[str, List[str]]] = None,
        ignore_dirs: Optional[List[str]] = None,
        ignore_case: bool = False,
        parent_logger=None,
        **kwargs
        ) -> List[Path]:
    """
    Gather all files in a directory.

    Parameters:
        directory (Union[str, Path]):
            The directory to gather files from.

        recursive (bool):
            Whether to gather files recursively.

        file_types (Optional[Union[str, List[str]]]):
            The file types (extensions, not including the leading '.') to gather. If None, files of all types will be
            gathered.

        ignore_dirs (Optional[List[str]]):
            A list of directory names to ignore. If any directory at any depth matches the given directory names it
            will be ignored. If `ignore_case` is True, the directory names will be converted to lowercase before
            comparison. This argument is ignored if :param:`recursive` is `False`.

        ignore_case (bool):
            Whether to ignore case when matching directory names. If True, the directory names will be converted to
            lowercase before comparison. This parameter is ignored if there are no directory names to ignore.

        parent_logger:
            The parent logger to use for the logger for this function.

    Returns:
        List[Path]:
            A list of file-paths for files in the directory.
    """
    from inspyre_toolbox.log_engine import ROOT_LOGGER
    MOD_LOGGER = ROOT_LOGGER.get_child('path_man')
    log = (parent_logger or MOD_LOGGER).get_child('gather_files_in_dir')
    from inspyre_toolbox.filesystem.file.helpers import get_file_object

    log.debug(f'Gathering files in directory: {directory}')

    directory = Path(directory).resolve()
    if not directory.is_dir():
        raise ValueError(f"Invalid directory: {directory}!")

    file_types = __normalize_file_types(file_types)
    ignore_dirs = __normalize_ignore_dirs(ignore_dirs, ignore_case)

    files = []
    for dir_path, dir_names, file_names in os.walk(directory):
        dir_names[:] = __filter_dirs(dir_names, ignore_dirs, ignore_case)

        if not recursive:
            dir_names.clear()

        files.extend(__filter_files(Path(dir_path), file_names, file_types, get_file_object))

    log.debug(f'Gathered {len(files)} files in directory: {directory} | Recursive: {recursive}')
    return files


def get_storage_unit_abbreviation(unit):
    """
    Get the abbreviation for a storage unit.

    Parameters:
        unit (str):
            The storage unit.

    Returns:
        str:
            The abbreviation for the storage unit.
    """
    return {
            'byte':      'B',
            'kilobyte':  'KB',
            'megabyte':  'MB',
            'gigabyte':  'GB',
            'terabyte':  'TB',
            'petabyte':  'PB',
            'exabyte':   'EB',
            'zetabyte':  'ZB',
            'yottabyte': 'YB'
            }.get(unit, 'B')



from inspyre_toolbox.exceptional import CustomRootException


class InspyreToolboxError(CustomRootException):
    """
    Base class for all Inspyre Toolbox errors.
    """
    default_message = 'An unknown error has occurred within the Inspyre Toolbox package.'

    def __init__(self, **kwargs):
        message = kwargs.pop('message', self.default_message)
        super().__init__(message, **kwargs)


class InvalidParameterCombinationError(InspyreToolboxError, ValueError):
    """
    Error raised when an invalid combination of parameters is provided.
    """
    default_message = 'An invalid combination of parameters was provided.'

    def __init__(self, **kwargs):
        message = kwargs.pop('message', self.default_message)
        InspyreToolboxError.__init__(self, message=message, **kwargs)


class MissingRequiredParameterError(InspyreToolboxError, ValueError):
    """
    Error raised when a required parameter is missing.
    """
    default_message = 'A required parameter is missing.'

    def __init__(self, **kwargs):
        message = kwargs.pop('message', self.default_message)
        InspyreToolboxError.__init__(self, message=message, **kwargs)



"""
This module contains classes and functions to instantiate and provide detailed package version information.

Classes:
    VersionParser:
        A class to parse and provide detailed version information.

    PyPiVersionInfo:
        A class to provide version information from PyPi.

Functions:
    None

Since:
    v1.6.0
"""

from inspyre_toolbox.ver_man.classes import PyPiVersionInfo, TestPyPiVersionInfo, RELEASE_MAP, VersionParser


__all__ = [
    'RELEASE_MAP',
    'VersionParser',
    'PyPiVersionInfo',
    'TestPyPiVersionInfo'
]



"""
This module provides helper functions for working with images.

Functions:
    get_supported_image_formats:
        Get the supported image formats.
"""

from pathlib import Path
from typing import Union

from platformdirs import user_pictures_path

from inspyre_toolbox.filesystem.file.collection import FileCollection, collect_files
from inspyre_toolbox.filesystem.file.images.constants import VALID_IMAGE_EXTENSIONS
from inspyre_toolbox.path_man import provision_path

DEFAULT_PICTURES_DIR = user_pictures_path()


def collect_image_files(root_dir: Union[str, Path] = DEFAULT_PICTURES_DIR, recursive: bool = True,
                        file_types=VALID_IMAGE_EXTENSIONS, **kwargs) -> FileCollection:
    return collect_files(root_dir, recursive, extensions=file_types, **kwargs)


def is_image_file(path: Union[str, Path], skip_path_provisioning=False) -> bool:
    """
    Check if a file is an image file.

    Parameters:
        path (Union[str, Path]):
            The path to the file. This can be a string or a Path object. If a string is provided, it will be converted
            to a Path object, expanded, resolved, and absolute.

        skip_path_provisioning (bool):
            Skip path provisioning if set to True. Default is False. If set to True, the path will not be
            provisioned. If this is set to True, the path must be a valid path, otherwise an exception will be raised.

    Returns:
        bool:
            True if the file is an image file, False otherwise.
    """
    if not skip_path_provisioning:
        path = provision_path(path)

    try:
        if not isinstance(path, Path):
            raise TypeError('The path must be a string or a Path object.')
    except TypeError as e:
        if skip_path_provisioning:
            print('If you\'re going to skip path provisioning, the path must be a Path object.')
        raise e from e

    return path.suffix[1:] in VALID_IMAGE_EXTENSIONS



"""
Contains the custom errors for the descriptors module.
"""


class UnmetConditionError(PermissionError):
    """
    An error raised when a condition is not met.
    """

    def __init__(self, message):
        """
        Initialize a new UnmetConditionError instance.

        Parameters:

            message (str):
                The error message.
        """
        super().__init__(message)



from argparse import ArgumentParser

from inspyre_toolbox.sys_man.operating_system.checks import is_linux

# Constants for program name and description
PROG = 'ist-add-to-path'
DESCRIPTION = 'Add a program path to the system PATH environment variable.'

LINUX = is_linux()

del is_linux


class Arguments(ArgumentParser):

    def __init__(self, *args, **kwargs):
        super().__init__(prog=PROG, description=DESCRIPTION, *args, **kwargs)

        # Define the command-line arguments for the program
        self.add_argument(
                'NEW PATH',
                type=str,
                help='The full path to the directory you want to add to the system PATH.'
                )

        if LINUX:
            self.add_argument(
                    '-f', '--file',
                    type=str,
                    help='The file to add the path to. Defaults to ~/.bashrc.',
                    required=False,
                    action='store'
                    )
        else:
            self.add_argument(
                    '-d', '--do-not-notify',
                    help='Do not notify the environment of changes.',
                    required=False,
                    action='store_true'
                    )

        self.__parsed = None

    @property
    def parsed(self):
        """
        Parse the command-line arguments and return the parsed arguments.

        Returns:
            Namespace:
                The parsed command-line arguments.

        """
        if not self.__parsed:
            self.__parsed = self.parse_args()
        return self.__parsed



from pathlib import Path
from typing import Optional, Union

from inspyre_toolbox.common.types import File as FileType
from inspyre_toolbox.console_kit.prompts.dialogs import ConfirmationPrompt
from inspyre_toolbox.decor import frozen_property
from inspyre_toolbox.filesystem import MOD_LOGGER as PARENT_LOGGER
from inspyre_toolbox.filesystem.file.helpers import get_file_checksum
from inspyre_toolbox.log_engine import Loggable
from inspyre_toolbox.path_man import provision_path
from inspyre_toolbox.syntactic_sweets.classes.decorators.type_validation import validate_type
from inspyre_toolbox.sys_man.operating_system.checks import is_windows

MOD_LOGGER = PARENT_LOGGER.get_child('file')


class File(Loggable, FileType):

    def __init__(
            self,
            path: Union[str, Path],
            auto_get_checksum: bool = False,
            backup_dir: Union[str, Path] = None
            ):
        """
        Initialize the File object.

        Parameters:
            path (Union[str, Path]):
                The path to the file.

            auto_get_checksum (bool):
                Whether to automatically get the checksum of the file.

            backup_dir (Path, optional):
                The directory to back up the file to. If `None`, the file will be backed up to a directory named
                'backups' in the parent directory.
        """
        # Start the logger.
        super().__init__(MOD_LOGGER)

        self.__auto_get_checksum = auto_get_checksum
        self.__extension = None
        self.__size = None
        self.__backup_dir = None

        self.__checksum = None

        self.__has_recall_attribute = None


        if is_windows():
            from inspyre_toolbox.filesystem.file.attributes import file_has_recall_attribute

            self.has_recall_attribute = file_has_recall_attribute(path, do_not_provision=True)

        self.path = path

        if self.auto_get_checksum:
            self.checksum = get_file_checksum(self.path)

        # If backup_dir is not provided, set it to a directory
        # named 'backups' in the parent directory.
        if backup_dir is not None:
            self.backup_dir = backup_dir
        else:
            self.backup_dir = self.path.parent.joinpath('backups')

    @property
    def auto_get_checksum(self) -> bool:
        """
        Get whether the file should automatically get the checksum.
        Returns:
            bool:
                Whether the file should automatically get the checksum.
        """
        return self.__auto_get_checksum

    @property
    def backup_dir(self) -> Path:
        """
        Get the backup directory for the file.

        Returns:
            Path:
                The backup directory for the file.
        """
        return self.__backup_dir

    @backup_dir.setter
    @validate_type(str, Path, preferred_type=Path, conversion_funcs={str: provision_path})
    def backup_dir(self, new: Union[str, Path]):
        """
        Set the backup directory for the file.

        Parameters:
            new (Union[str, Path]):
                The new backup directory for the file.

        Returns:
            None
        """
        self.__backup_dir = Path(new)

    @property
    def backup_path(self) -> Path:
        """
        Get the backup path for the file.

        Note:
            This is a read-only property, and depends on :attr:`path` and :attr:`backup_dir`.

        Returns:
            Path:
                The backup path for the file.
        """
        return self.backup_dir / self.path.name

    @property
    def checksum(self) -> str:
        """
        Return the file's checksum.

        Returns:
            str:
                The file's checksum.

        Raises:
            FileNotFoundError:
                If the file is not found.
        """
        return self.__checksum

    @checksum.setter
    @frozen_property('checksum', allowed_types=str, restrict_setter=True)
    def checksum(self, new: str):
        """
        Set the file's checksum. This is a frozen property, and cannot be changed once set.

        Parameters:
            new (str):
                The new checksum for the file.

        Returns:
            None
        """
        self.__checksum = new

    @property
    def exists(self) -> bool:
        """
        Check if the file exists.

        Returns:
            bool:
                True if the file exists, False otherwise.
        """
        return self.path.exists()

    @property
    def extension(self) -> str:
        """
        Get the extension of the file.

        Returns:
            str:
                The extension of the file.
        """
        return self.__extension

    @property
    def has_recall_attribute(self) -> bool:
        """
        Check if the file has the recall attribute.

        Returns:
            bool:
                True if the file has the recall attribute, False otherwise.
        """
        if not is_windows():
            raise NotImplementedError("The 'has_recall_property' attribute is only available on Windows.")

        return self.__has_recall_attribute

    @has_recall_attribute.setter
    @frozen_property('has_recall_attribute', allowed_types=bool, restrict_setter=True)
    def has_recall_attribute(self, new):
        """
        Set whether the file has the recall attribute. This is a frozen property, and cannot be changed once set.

        Parameters:
            new (bool):
                Whether the file has the recall attribute.

        Returns:
            None
        """
        self.__has_recall_attribute = new

    @property
    def is_image(self):
        """
        Check if the file is an image.

        Returns:
            bool:
                True if the file is an image, False otherwise.
        """
        from inspyre_toolbox.filesystem.file.images.helpers import is_image_file

        return is_image_file(self.path)

    @property
    def is_local(self) -> bool:
        """
        Check if the file is local.

        Returns:
            bool:
                True if the file is local, False otherwise.
        """
        try:
            non_local = self.has_recall_attribute
            return not non_local
        except NotImplementedError:
            return True

    @property
    def name(self) -> str:
        """
        Get the name of the file.

        Returns:
            str:
                The name of the file.
        """
        return self.path.name

    @property
    def path(self) -> Path:
        """
        Get the path of the file.

        Returns:
            Path:
                The path of the file.
        """

        return Path(self.__path)

    @path.setter
    @frozen_property('path', allowed_types=(Path, str), restrict_setter=False)
    @validate_type(str, Path, preferred_type=Path, conversion_funcs={str: provision_path})
    def path(self, new: Union[str, Path]):
        """
        Set the path of the file.

        Args:
            new:

        Returns:

        """
        self.__path = Path(new)
        self.__extension = self.__path.suffix
        self.__size = self.__path.stat().st_size

    @property
    def size_in_bytes(self) -> int:
        """
        Get the size of the file in bytes.

        Returns:
            int:
                The size of the file in bytes.
        """
        return self.__size

    def back_up(self, backup_dir=None, backup_name=None, backup_extension=None, overwrite=False):
        """
        Back up the file.

        Parameters:

            backup_dir (Path, optional):
                The directory to back up the file to. If `None`, the :attr:`backup_dir` will be used.

            backup_name (str, optional):
                The name of the backup file. If `None`, the name of the file will be used.

            backup_extension (str, optional):
                The extension of the backup file. If `None`, the extension of the file will be used.

            overwrite (bool):
                Whether to overwrite the backup file if it already exists. If `False`, a :class:`FileExistsError` will
                be raised.

        Returns:
            None

        Raises:

            FileExistsError:
                If the backup file already exists and `overwrite` is `False`.
        """
        log = self.create_logger()
        log.debug('Backing up file.')

        if backup_dir is None:
            backup_dir = self.backup_dir

        if backup_name is None:
            backup_name = self.path.stem

        if backup_extension is None:
            backup_extension = self.extension

        backup_path = backup_dir / f'{backup_name}{backup_extension}'

        if backup_path.exists() and not overwrite:
            raise FileExistsError(f'The backup file already exists: {backup_path}')

        log.debug(f'Copying file to: {backup_path}')
        self.path.replace(backup_path)

    def copy(
            self,
            destination_dir,
            new_name=None,
            new_extension=None,
            overwrite=None,
            skip_confirm_on_non_local=False,
            skip_confirm_on_overwrite=False,
            confirmation_prompt=None
            ):
        """
        Copy the file to the specified directory.

        Parameters:
            destination_dir (Path):
                The directory to copy the file to.

            new_name (str, optional):
                The new name of the file.

            new_extension (str, optional):
                The new extension of the file.

            overwrite (bool):
                Whether to overwrite the file if it already exists in the destination directory. The user will
                be prompted for confirmation, unless `skip_confirm_on_overwrite` is `True`.

            skip_confirm_on_non_local (bool):
                Whether to skip the confirmation if the file is not local.

            skip_confirm_on_overwrite (bool):
                Whether to skip the confirmation if the file already exists in the destination directory.

        Returns:
            None

        Raises:
            FileExistsError:
                If the file already exists in the destination directory and `overwrite` is `False`.
        """
        log = self.create_logger()
        log.debug(f'Copying file to: {destination_dir}')
        if self.has_recall_attribute:
            recall_prompt_text = ('The file is not local. Copying would require downloading the file. Are you sure you '
                                  'want to '
                                  f'proceed? \n\nDestination: {destination_dir}\n\nFile: {self.path}\n\nSize: {self.size_in_bytes} bytes')

            if not skip_confirm_on_non_local:
                if confirmation_prompt is None:
                    confirmation_prompt = ConfirmationPrompt(
                            title='Non-Local File',
                            text='The file is not local. Are you sure you want to proceed?',
                            no_text_stub=True,
                            no_title_stub=True
                            )

                confirmation_prompt.run()

                if not confirmation_prompt.answer:
                    denied_statement = 'User did not confirm to proceed with the download.'

                    log.debug(denied_statement)
                    log.error(denied_statement)
                    return

                log.debug('User confirmed to proceed with the download.')

        if new_name is None:
            new_name = self.path.stem

        if new_extension is None:
            new_extension = self.extension

        destination_path = provision_path(destination_dir / f'{new_name}{new_extension}')

        if destination_path.exists() and not overwrite and not skip_confirm_on_overwrite:
            raise FileExistsError(f'The file already exists in the destination directory: {destination_path}')

        log.debug(f'Copying file to: {destination_path}')
        self.path.replace(destination_path)

    def move(self, destination_dir, new_name=None, new_extension=None, overwrite=False):
        """
        Move the file to the specified directory.

        Parameters:
            destination_dir (Path):
                The directory to move the file to.
            new_name (str, optional):
                The new name of the file.
            new_extension (str, optional):
                The new extension of the file.
            overwrite (bool):
                Whether to overwrite the file if it already exists in the destination directory.

        Returns:
            None

        Raises:
            FileExistsError:
                If the file already exists in the destination directory and `overwrite` is `False`.
        """
        log = self.create_logger()
        log.debug(f'Moving file to: {destination_dir}')

        if new_name is None:
            new_name = self.path.stem

        if new_extension is None:
            new_extension = self.extension

        destination_path = provision_path(destination_dir / f'{new_name}{new_extension}')

        if destination_path.exists() and not overwrite:
            raise FileExistsError(f'The file already exists in the destination directory: {destination_path}')

        log.debug(f'Moving file to: {destination_path}')
        self.path.replace(destination_path)

    def get_checksum(
            self,
            recalculate=False,
            skip_confirm_on_non_local=False,
            confirmation_prompt=None,
            error_on_fail=False
            ) -> Optional[str]:
        """
        Get the checksum of the file.

        Parameters:
            recalculate (bool):
                Whether to recalculate the checksum.
            skip_confirm_on_non_local (bool):
                Whether to skip the confirmation if the file is not local.
            confirmation_prompt (ConfirmationPrompt, optional):
                The confirmation prompt to use if the file is non-local.
            error_on_fail (bool):
                Whether to raise an error if the user does not confirm.

        Returns:
            str:
                The checksum of the file.
        """
        log = self.create_logger()
        non_local = not self.is_local
        log.debug(f'File is non-local: {non_local}')

        if non_local and not skip_confirm_on_non_local:
            if confirmation_prompt is None:
                confirmation_prompt = ConfirmationPrompt(
                        title='Non-Local File',
                        text='The file is not local. Are you sure you want to proceed?',
                        no_text_stub=True,
                        no_title_stub=True
                        )

            confirmation_prompt.run()

            if not confirmation_prompt.answer:
                denied_statement = 'User did not confirm to proceed with the download.'

                log.debug(denied_statement)
                log.error(denied_statement)

                if error_on_fail:
                    raise PermissionError(denied_statement)
                return None

            log.debug('User confirmed to proceed with the download.')

        if recalculate or not self.checksum:
            self.checksum = get_file_checksum(self.path)

        return self.checksum

    def get_extension(self):
        """
        Get the extension of the file.

        Returns:
            str:
                The extension of the file.
        """
        return self.path.suffix


def non_local_confirmation_prompt(file: File, operation=None, text=None) -> ConfirmationPrompt:
    """
    Create a confirmation prompt for a non-local file.

    Parameters:
        file (File):
            The file to create the prompt for.

        operation (str, optional):
            The operation to perform on the file. If provided, the prompt will include the operation in the text.

        text (str, optional):
            The text to include in the prompt. If provided, the operation will be ignored.

    Returns:
        ConfirmationPrompt:
            The confirmation prompt for the non-local file.
    """
    operations = {
            'checksum': 'Calculating the checksum for',
            'copy':     'copying',
            'move':     'moving',
            'backup':   'backing-up'
            }

    title = 'Non-Local File'

    if operation is None and text is None:
        raise ValueError('Either `operation` or `text` must be specified.')

    if text is None and operation in operations:
        text = (f'The file is not local. {operations[operation.lower()]} the file would require downloading the it. '
                'Are you sure you want to proceed?\n\n'
                f'Operation: {operations[operation]}\nFile: {file.path}\nSize: {file.size_in_bytes} bytes')

    prompt = ConfirmationPrompt(title=title, text=text, no_text_stub=True, no_title_stub=True)

    return prompt



import json
from typing import Dict, Union

from inspyre_toolbox.config_man.options.shortcuts import Option, StringInputOption, YesNoOption


class OptionManager:
    """
    Manages various options, acting like a dictionary with type enforcement.
    Tracks changes through the OptionLedger.
    """

    def __init__(self):
        self._options = {}

    @property
    def options(self):
        return self._options

    def add_option(self, option: Option):
        """Add an option to the manager."""
        if option.name in self._options:
            raise ValueError(f"Option '{option.name}' already exists.")
        self._options[option.name] = option

    def set_option_value(self, name: str, value: Union[bool, str]):
        """Set the value of an option by name."""
        if option := self._options.get(name):
            option.set_value(value)
        else:
            raise KeyError(f"Option '{name}' not found.")

    def __getitem__(self, key: str) -> Union[bool, str]:
        """Allow dictionary-style access to get option values."""
        return self._options[key].value

    def __setitem__(self, key: str, value: Union[bool, str]):
        """Allow dictionary-style access to set option values."""
        self.set_option_value(key, value)

    def __contains__(self, key: str) -> bool:
        """Check if an option exists by key."""
        return key in self._options

    def __repr__(self):
        """Return a string representation of all options."""
        return "\n".join([f"{name}: {option}" for name, option in self._options.items()])

    def save_to_dict(self) -> Dict[str, Dict]:
        """Convert all options to a dictionary that can be serialized."""
        return {name: option.to_dict() for name, option in self._options.items()}

    def load_from_dict(self, data: Dict[str, Dict]):
        """Load options from a dictionary."""
        for name, option_data in data.items():
            option_type = option_data.pop('type')
            if option_type == 'YesNoOption':
                self._options[name] = YesNoOption.from_dict(option_data)
            elif option_type == 'StringInputOption':
                self._options[name] = StringInputOption.from_dict(option_data)

    def save_to_file(self, filename: str):
        """Save all options to a file."""
        with open(filename, 'w') as file:
            json.dump(self.save_to_dict(), file, indent=4)

    def load_from_file(self, filename: str):
        """Load options from a file."""
        with open(filename, 'r') as file:
            data = json.load(file)
            self.load_from_dict(data)

    def get_ledger_entries(self):
        """Get the ledger entries for all tracked changes."""
        if Option.ledger is None:
            return "No changes have been recorded."
        return Option.ledger



import base64

from inspyre_toolbox.filesystem.file.images.constants import *

__all__ = [
        'ImageFile',
        'VALID_IMAGE_EXTENSIONS',
        'IMAGE_FORMATS',
        'PIL_EXTENSIONS',
        'PIL_FORMATS',
        ]

from inspyre_toolbox.filesystem.file import File
from inspyre_toolbox.filesystem.file.images.helpers import is_image_file


class ImageFile(File):
    """
    A class for handling image files.
    """

    def __init__(self, path):
        """
        Initialize the ImageFile object.

        Args:
            path (str): The path to the image file.
        """
        super().__init__(path)

        self._format = self.get_format()

        if not self.is_valid():
            raise ValueError(f'Invalid image file: {self.path}')

    def get_base64(self):
        """
        Get the base64 encoded image data.

        Returns:
            str: The base64 encoded image data.
        """
        with open(self.path, 'rb') as file:
            return base64.b64encode(file.read()).decode('utf-8')

    def get_format(self):
        """
        Get the format of the image file.

        Returns:
            str: The format of the image file.
        """
        return self.get_extension()

    def is_valid(self):
        """
        Check if the image file is valid.

        Returns:
            bool: True if the image file is valid, False otherwise.
        """
        return is_image_file(self.path)



import hashlib
from pathlib import Path
from typing import List, Optional, Union

from inspyre_toolbox.conversions.bytes import ByteConverter
from inspyre_toolbox.path_man import provision_path


def __normalize_file_types(file_types: Optional[Union[str, List[str]]]) -> List[str]:
    """Normalize file types to a list of strings."""
    if file_types is None:
        return ['*']

    return [file_types] if isinstance(file_types, str) else file_types


def __normalize_ignore_dirs(ignore_dirs: Optional[List[str]], ignore_case: bool) -> set:
    """Normalize and prepare directory names to ignore."""
    if ignore_case and ignore_dirs:
        return {dir_name.lower() for dir_name in ignore_dirs}
    return set(ignore_dirs or [])


def __filter_dirs(dir_names: List[str], ignore_dirs: set, ignore_case: bool) -> List[str]:
    """Filter out directories that should be ignored."""
    if ignore_case:
        return [d for d in dir_names if d.lower() not in ignore_dirs]
    return [d for d in dir_names if d not in ignore_dirs]


def __filter_files(dir_path: Path, file_names: List[str], file_types: List[str], get_file_object) -> List[Path]:
    """Filter and gather files based on file types."""
    return [
            get_file_object(dir_path / file_name)
            for file_name in file_names
            if any(file_name.endswith(f".{ftype}") for ftype in file_types) or '*' in file_types
            ]



def get_file_checksum(file_path, algorithm='sha256'):
    """
    Calculate the checksum of a file using the specified hashing algorithm.

    Parameters:
    file_path (str): Path to the file.
    algorithm (str): Hashing algorithm to use ('md5', 'sha1', 'sha256', etc.). Default is 'sha256'.

    Returns:
    str: The calculated checksum.

    Raises:
    ValueError: If the specified algorithm is not supported.
    FileNotFoundError: If the specified file does not exist.
    """
    # Validate the algorithm
    if algorithm not in hashlib.algorithms_available:
        raise ValueError(f"Unsupported algorithm '{algorithm}'. Available algorithms: {hashlib.algorithms_available}")

    # Initialize the hash object
    hash_func = hashlib.new(algorithm)

    # Read the file in chunks and update the hash object
    try:
        with open(file_path, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_func.update(chunk)
    except FileNotFoundError as e:
        raise FileNotFoundError(f"The file '{file_path}' does not exist.") from e

    # Return the hex digest of the hash
    return hash_func.hexdigest()


def get_lowest_unit_size(size: int) -> tuple[Union[int, float], str]:
    """
    Get the lowest unit size for a given size.

    Parameters:
        size (int):
            The size to convert.

    Returns:
        tuple[Union[int, float], str]:
            The converted size and the unit.

    Examples:
        >>> get_lowest_unit_size(1024)
        (1.0, 'KB')
        >>> get_lowest_unit_size(1024 * 1024)
        (1.0, 'MB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024)
        (1.0, 'GB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024)
        (1.0, 'TB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'PB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'EB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'ZB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'YB')
    """
    units = ['byte', 'kilobyte', 'megabyte', 'gigabyte', 'terabyte', 'petabyte', 'exabyte', 'zetabyte', 'yottabyte']
    units.reverse()
    converter = ByteConverter(size, 'byte')

    for unit in units:
        converted = converter.convert(unit.lower())

        if converted >= 1:
            return converted, unit.upper(),


def get_file_object(file_path, skip_path_provision=False):
    """
    Get a file object for the specified file path.

    Parameters:
        file_path (Union[str, Path]):
            The path to the file.

        skip_path_provision (bool):
            Skip path provisioning if set to True. Default is False.

    Returns:
        Optional[File, ImageFile]:
            A file object for the specified file path.

    Raises:
        ValueError:
            If the specified file path is not a valid file.
    """
    from inspyre_toolbox.filesystem.file import File
    from inspyre_toolbox.filesystem.file.images import ImageFile
    from inspyre_toolbox.filesystem.file.images.helpers import is_image_file

    if isinstance(file_path, Path):
        file_path = str(file_path)

    file_path = file_path if skip_path_provision else provision_path(file_path)

    return ImageFile(file_path) if is_image_file(file_path) else File(file_path)


def get_path_list_from_list_of_file_objects(file_objects):
    """
    Get a list of file paths from a list of file objects.

    Parameters:
        file_objects (List[File, ImageFile]):
            A list of file objects.

    Returns:
        List[str]:
            A list of file paths.
    """
    return [file_object.path for file_object in file_objects]



"""
This module provides a class to convert values between different byte units.

Classes:
    ByteConverter

Functions:
    get_lowest_unit_size:
        Get the lowest unit size for a given size.

Variables:
    VERSION:
        The version of the module.

    MOD_LOGGER:
        The logger for the module.

    MOD_LOGGER_NAME:
        The name of the logger for the module.


ByteConverter:
    A class to convert values between different byte units.

    Attributes:
        units (dict):
            A dictionary holding the conversion units.

        BYTE_FAMILY (list):
            A list of byte units.

        BIT_FAMILY (list):
            A list of bit units.

    Methods:
        convert(unit: str) -> float:
            Converts the bytes value to the specified unit.

        get_lowest_safe_conversion(keep_family: bool) -> tuple(float, str):
            Returns the lowest safe conversion for the value, a tuple containing the value and unit,
            respectively.

    Properties:
        conversion_map:
            Returns the conversion map.

        family_factors:
            Returns the factors of the family of the initial unit.

        keep_family:
            Returns the keep_family attribute.

        initial_bytes:
            Returns the initial value in bytes.

        initial_unit:
            Returns the initial unit of the value.

        initial_value:
            Returns the initial value.


"""
from typing import Any, Union

from inspy_logger import InspyLogger, Loggable

from inspyre_toolbox.common.about.version import VERSION_NUMBER
from inspyre_toolbox.syntactic_sweets.classes import validate_type

VERSION = VERSION_NUMBER

MOD_LOGGER = InspyLogger('inspyre_toolbox.conversions.bytes')

MOD_LOGGER.debug(f'Byte conversion module loaded. | Inspyre-Toolbox v{VERSION}')


class ByteConverter(Loggable):
    """
    ByteConverter class for converting between different units of digital information storage.

    This class provides a flexible way to convert between different byte and bit units
    such as bytes, kilobytes, megabytes, gigabytes, and their respective bit equivalents.

    Attributes:
    ----------
    value (float):
        The numeric value of the data amount.

    unit (str):
        The unit of the data amount, e.g., 'byte', 'kilobyte', etc.

    UNIT_FAMILIES (list):
        A dictionary that organizes units into families based on their scale, e.g., byte-based units.

    UNIT_CONVERSIONS (dict):
        A dictionary that maps unit strings to their conversion factors relative to 1 byte.

    Methods:
    -------
    convert(to_unit, strict_case=True):
        Converts the stored value to a specified unit.

    get_lowest_safe_conversion(keep_family=True):
        Returns the smallest unit (in terms of scale) that represents the value without falling below 1 in the desired family.
    """
    # Dictionary holding the conversion units
    UNIT_CONVERSIONS = {
        'bit': 1/8,
            'byte':      1,
            'kilobit':   125,
            'kilobyte':  1000,
            'megabit':   125000,
            'megabyte':  1000000,
            'gigabit':   125000000,
            'gigabyte':  1000000000,
            'terabit':   125000000000,
            'terabyte':  1000000000000,
            'petabit':   125000000000000,
            'petabyte':  1000000000000000,
            'exabit':    125000000000000000,
            'exabyte':   1000000000000000000,
            'zetabit':  125000000000000000000,
            'zetabyte': 1000000000000000000000,
            'yottabit':  125000000000000000000000,
            'yottabyte': 1000000000000000000000000
            }

    UNIT_FAMILIES = {
            'byte': ['byte', 'kilobyte', 'megabyte', 'gigabyte', 'terabyte', 'petabyte', 'exabyte', 'zetabyte',
                     'yottabyte'],
            'bit':  ['bit', 'kilobit', 'megabit', 'gigabit', 'terabit', 'petabit', 'exabit', 'zetabit', 'yottabit']
            }

    def __init__(
            self,
            value: Union[int, float],
            unit: str
            ):
        """
        Initialize a ByteConverter object with a specific value and unit.

        Parameters:
        ----------
        value (Union[int, float]):
            The numeric value of the data amount.

        unit : str
            The unit of the data amount, e.g., 'byte', 'kilobyte', etc.
        """
        super().__init__(parent_log_device=MOD_LOGGER)
        # self.__keep_family = None
        #
        # unit = unit.removesuffix('s')
        #
        # if unit not in self.UNITS:
        #     raise ValueError(f'Invalid unit: {unit}')
        #
        # if value < 0:
        #     raise ValueError('Value cannot be negative.')
        #
        # self.__initial_unit = unit
        # self.__initial_value = value
        #
        # # Convert the input value to bytes
        # self.__bytes = value * self.UNITS[unit]
        # self.keep_family = keep_family

        unit = unit.lower()
        self.__initial_value = value
        self.__initial_unit = unit

        self.__value = None

        self.unit = unit.lower()

        self.__family_factors = self.__calculate_family_factors()

    def __calculate_family_factors(self) -> dict:
        """
        Calculate the family factors based on the UNIT_FAMILIES and UNIT_CONVERSIONS.

        Returns:
        -------
        dict:
            A dictionary of family units and their conversion factors.
        """
        family_factors = {}

        for family, units in self.UNIT_FAMILIES.items():
            family_factors[family] = {unit: self.UNIT_CONVERSIONS[unit] for unit in units}

        return family_factors

    def convert(self, to_unit, strict_case=True):
        """
        Convert the stored value to a specified unit.


        Parameters:
        ----------
        to_unit(str):
            The target unit for conversion.

        strict_case(Optional[bool])
            Whether to strictly enforce case sensitivity (default is True).

        Returns:
        -------
        float:
            The converted value.
        """
        log = self.create_child_logger()

        if not strict_case:
            log.debug(f'Lowering the case of the target unit; {to_unit}')
            to_unit = to_unit.lower()
        else:
            log.debug(f'Strict case sensitivity is enabled. Target unit: {to_unit}')

        if to_unit not in self.UNIT_CONVERSIONS:
            log.error(f"Invalid unit: {to_unit}. Valid units are: {list(self.UNIT_CONVERSIONS.keys())}")
            raise ValueError(f"Invalid unit: {to_unit}. Valid units are: {list(self.UNIT_CONVERSIONS.keys())}")

        from_factor = self.UNIT_CONVERSIONS[self.unit]

        to_factor = self.UNIT_CONVERSIONS[to_unit]

        log.debug(f"To factor: {to_factor} | From factor: {from_factor}")

        converted_value = self.value * from_factor / to_factor
        log.debug(f"Converted {self.value} {self.unit} to {converted_value} {to_unit}")

        return converted_value

    def get_lowest_safe_conversion(
            self,
            keep_family: bool = False
            ) -> tuple[Any, int | float]:
        """
        Returns the lowest safe conversion for the value.

        Parameters:
            keep_family (bool):
                A flag indicating whether to keep the family of the initial unit.

        Returns:
            tuple:
                The lowest safe conversion value and unit.
        """
        log_name = f'{self.log_device.name}:get_lowest_safe_conversion'
        log = self.log_device.find_child_by_name(log_name)[0] if self.log_device.has_child(
            log_name) else self.create_child_logger()

        sorted_units = None

        if keep_family:
            log.debug(f'Keeping the family of the initial unit: {self.__initial_unit}')
            for family, units in self.UNIT_FAMILIES.items():
                if self.unit in units:
                    sorted_units = sorted(units, key=lambda x: self.UNIT_CONVERSIONS[x], reverse=True)
                    break
        else:
            sorted_units = sorted(self.UNIT_CONVERSIONS.keys(), key=lambda x: self.UNIT_CONVERSIONS[x], reverse=True)

        if sorted_units is None:
            log.error(f'No units found for the initial unit: {self.__initial_unit}')
            raise ValueError(f'No units found for the initial unit: {self.__initial_unit}')

        for unit in sorted_units:
            converted_value = self.convert(unit)
            if converted_value >= 1:
                log.debug(
                    f'Lowest safe conversion: {converted_value} {unit} | Original value: {self.__initial_value} {self.__initial_unit}')
                return unit, converted_value

        return self.unit, self.value

        #     family = self.BYTE_FAMILY if self.__initial_unit in self.BYTE_FAMILY else self.BIT_FAMILY
        #     log.debug(f'Family of the initial unit: {family}')
        #     factors = [self.UNITS[unit] for unit in family]
        # else:
        #     log.debug('Not keeping the family of the initial unit.')
        #     factors = [self.UNITS[unit] for unit in self.UNITS]
        #
        # log.debug(f'Factors: {factors}')
        #
        # base_value = self.__initial_value * self.UNITS[self.__initial_unit]
        # log.debug(f'Base value: {base_value}')
        #
        # lowest_value = float('inf')
        # lowest_unit = None
        #
        # for unit, factor in self.UNITS.items():
        #     log.debug(f'Checking unit: {unit}')
        #     log.debug(f'Factor: {factor}')
        #     if factor in factors:
        #         log.debug(f'Factor in factors: {factor}')
        #         converted_value = base_value / factor
        #         log.debug(f'Converted value: {Numerical(converted_value, unit).count_noun()}')
        #         if 1.0 <= converted_value < lowest_value:
        #             log.debug(f'Converted value is less than lowest value: {converted_value}')
        #             lowest_value = converted_value
        #             log.debug(f'New lowest value: {lowest_value}')
        #             lowest_unit = unit
        #             log.debug(f'New lowest unit: {lowest_unit}')
        #
        # log.debug(f'Lowest value: {lowest_value}, Lowest unit: {lowest_unit}')
        #
        # if lowest_unit is None:
        #     log.debug('No unit found. Finding the smallest factor.')
        #     smallest_factor_index = factors.index(min(factors))
        #     log.debug(f'Smallest factor index: {smallest_factor_index}')
        #     lowest_unit = list(self.UNITS.keys())[smallest_factor_index]
        #     log.debug(f'Lowest unit: {lowest_unit}')
        #     lowest_value = base_value / min(factors)
        #     log.debug(f'Lowest value: {Numerical(lowest_value, lowest_unit).count_noun()}')
        #
        # log.debug(f'Final lowest value: {lowest_value}, Final lowest unit: {lowest_unit} | '
        #           f'{Numerical(lowest_value, lowest_unit).count_noun()}')
        #
        # return lowest_value, lowest_unit

    @property
    def family_factors(self) -> list[float]:
        """
        Returns the factors of the family of the initial unit.

        Returns:
            list:
                The factors of the family of the initial unit.

        """
        return self.__family_factors

    @property
    def keep_family(self):
        """
        Returns the keep_family attribute.

        Returns:
            bool:
                The keep_family attribute.
        """
        return self.__keep_family

    @keep_family.setter
    @validate_type(bool)
    def keep_family(self, new):
        """
        Sets the keep_family attribute.

        Parameters:
            new (bool):
                The value to set the keep_family attribute to.

        Returns:
            None
        """
        self.__keep_family = new

    @property
    def initial_bytes(self):
        """
        Returns the initial value in bytes.

        Returns:
            float:
                The initial value in bytes.
        """
        return self.__bytes

    @property
    def initial_unit(self) -> str:
        """
        Returns the initial unit of the value.

        Returns:
            str:
                The initial unit of the value
        """
        return self.__initial_unit

    @initial_unit.setter
    def initial_unit(self, unit):
        """
        Placeholder for the initial unit setter.

        This setter is a placeholder and raises an AttributeError when called.

        Parameters:
            unit:
                Unused parameter.

        Returns:
            None

        Raises:
            AttributeError:
                The initial unit cannot be changed.
        """
        raise AttributeError("The initial unit cannot be changed.")

    @property
    def initial_value(self):
        """
        Returns the initial value.

        Returns:
            Union[int, float]:
                The initial value.
        """
        return self.__initial_value

    @initial_value.setter
    def initial_value(self, value):
        """
        Placeholder for the initial value setter.

        This setter is a placeholder and raises an AttributeError when called.

        Parameters:
            value:
                Unused parameter.

        Returns:
            None

        Raises:
            AttributeError:
                The initial value cannot be changed.
        """
        raise AttributeError("The initial value cannot be changed.")

    @property
    def units(self):
        return self.UNITS

    @property
    def value(self):
        return self.__value or self.__initial_value

    @value.setter
    @validate_type(int, float, preferred_type=float)
    def value(self, new):
        self.__value = new

    # Below are property methods that return the value in a specific unit

    @property
    def bits(self):
        """Returns the value in bits"""
        return self.convert('bit')

    @property
    def kilobits(self):
        """Returns the value in kilobits"""
        return self.convert('kilobit')

    @property
    def megabits(self):
        """Returns the value in megabits"""
        return self.convert('megabit')

    @property
    def gigabits(self):
        """Returns the value in gigabits"""
        return self.convert('gigabit')

    @property
    def terabits(self):
        """Returns the value in terabits"""
        return self.convert('terabit')

    @property
    def petabits(self):
        """Returns the value in petabits"""
        return self.convert('petabit')

    @property
    def exabits(self):
        """Returns the value in exabits"""
        return self.convert('exabit')

    @property
    def zetabits(self):
        """Returns the value in zetabits"""
        return self.convert('zetabit')

    @property
    def yottabits(self):
        """Returns the value in yottabits"""
        return self.convert('yottabit')

    @property
    def bytes(self):
        """Returns the value in bytes"""
        return self.convert('byte')

    @property
    def kilobytes(self):
        """Returns the value in kilobytes"""
        return self.convert('kilobyte')

    @property
    def megabytes(self):
        """Returns the value in megabytes"""
        return self.convert('megabyte')

    @property
    def gigabytes(self):
        """Returns the value in gigabytes"""
        return self.convert('gigabyte')

    @property
    def terabytes(self):
        """Returns the value in terabytes"""
        return self.convert('terabyte')

    @property
    def petabytes(self):
        """Returns the value in petabytes"""
        return self.convert('petabyte')

    @property
    def exabytes(self):
        """Returns the value in exabytes"""
        return self.convert('exabyte')

    @property
    def zetabytes(self):
        """Returns the value in zetabytes"""
        return self.convert('zetabyte')

    @property
    def yottabytes(self):
        """Returns the value in yottabytes"""
        return self.convert('yottabyte')

    def __str__(self):
        return f'{self.__initial_value} {self.__initial_unit}'

    def __repr__(self):
        return \
            (
                    f'<ByteConverter@{hex(id(self))}: '
                    f'Initial value: {self.__initial_value} | Initial unit: {self.__initial_unit}>'
            )


def get_lowest_unit_size(size: int) -> tuple[Union[int, float], str]:
    """
    Get the lowest unit size for a given size.

    Parameters:
        size (int):
            The size to convert.

    Returns:
        tuple[Union[int, float], str]:
            The converted size and the unit.

    Examples:
        >>> get_lowest_unit_size(1024)
        (1.0, 'KB')
        >>> get_lowest_unit_size(1024 * 1024)
        (1.0, 'MB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024)
        (1.0, 'GB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024)
        (1.0, 'TB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'PB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'EB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'ZB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'YB')
    """
    units = ['byte', 'kilobyte', 'megabyte', 'gigabyte', 'terabyte', 'petabyte', 'exabyte', 'zetabyte', 'yottabyte']
    units.reverse()
    converter = ByteConverter(size, 'byte')

    for unit in units:
        converted = converter.convert(unit.lower())

        if converted >= 1:
            return converted, unit.upper(),



import winreg
from warnings import warn

from inspyre_toolbox.path_man import prepare_path


def notify_environment_of_path_changes():
    """
    Notify the environment of PATH changes.
    """
    import ctypes

    # Notify the environment of changes
    ctypes.windll.user32.SendMessageTimeoutW(
            winreg.HWND_BROADCAST,
            winreg.WM_SETTINGCHANGE,
            0,
            "Environment",
            winreg.SMTO_ABORTIFHUNG,
            5000,
            None
            )


def add_to_path(new_path: str, do_not_notify: bool = False, profile_file=None):
    """
    The add_to_path function adds the program path to the system's PATH variable on Windows systems.
    This allows for easy access to programs from any terminal window.

    Args:
        new_path (str):
            The full path to the directory you'd like to add to your system's PATH
            environment variable.

    Returns:
        None
    """
    if profile_file:
        warn("The profile_file parameter is not used on Windows systems. Ignoring parameter.")

    update_user_path_variable(new_path)

    if not do_not_notify:
        notify_the_environment_of_changes()
        return print(f"Added {new_path} to path on Windows, changes have been notified to the environment")

    print(f"Added {new_path} to path on Windows, please restart shell for changes to take effect")


def update_user_path_variable(new_path: str):
    """
    The update_user_path_variable function updates the user's PATH variable with the new program path.

    Note:
        This function is for Windows systems only. It will not work on *nix systems.

        This function will not notify the environment of changes. *You must call notify_the_environment_of_changes
        after calling this function to notify the environment of changes*.

    Parameters:
        new_path (str):
            The full path to the directory you'd like to add to your PATH environment variable.

    Returns:
        None
    """
    new_path = prepare_path(new_path, do_not_create=True)
    with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as root:  # Get the current user
        with winreg.OpenKey(root, "Environment", 0, winreg.KEY_ALL_ACCESS) as key:  # Go to the environment key
            existing_path_value = winreg.EnumValue(key, 3)[1]  # Grab the current path value

            # Takes the current path value and appends the new program path
            new_path_value = existing_path_value + new_path + ";"

            # Update path variable with the new path value
            winreg.SetValueEx(key, "PATH", 0, winreg.REG_EXPAND_SZ, new_path_value)



from typing import Optional

from rich.console import Console
from rich.panel import Panel
from rich.text import Text


class CustomRootException(Exception):
    """Base class for exceptions in the CustomRoot module."""
    PANEL = Panel
    PANEL_TEXT = Text

    default_message = 'An unknown error has occurred.'

    def __init__(self, message: Optional[str] = None, skip_print: bool = False):
        """

        Args:
            message:
        """
        super().__init__(message or self.default_message)
        self.__message = None
        self.__printed = False
        self._skip_print = None
        self.message = message or self.default_message

        self.skip_print = skip_print

        if not self.skip_print:
            self.print_rich_panel()

    @property
    def message(self):
        additional_message = ''
        if hasattr(self, 'additional_message') and self.additional_message:
            additional_message = f'\n\n[Additional Information]:\n    {self.additional_message}\n\n'
        return f'\n\n  {self.__message}{additional_message}'

    @message.setter
    def message(self, new):
        if self.__message:
            return

        if not isinstance(new, str):
            raise ValueError('message must be a string.')

        self.__message = new.strip()

    @property
    def printed(self):
        return self.__printed

    @property
    def skip_print(self):
        return self._skip_print

    @skip_print.setter
    def skip_print(self, new):
        if not isinstance(new, bool):
            raise ValueError('skip_print must be a boolean.')

        if not self.printed:
            self._skip_print = new

    def print_rich_panel(self):
        console = Console()
        console.print(self.__rich__())
        self.__printed = True

    def __rich__(self):
        return Panel(
            Text(self.message[1:], style='bold red'),
                title=self.__class__.__name__,
                style='bold red',
                )

    # Create a __str__ method that returns the rich rendering
    def __str__(self):
        return self.message



"""
Package information for Inspyre-Toolbox.

This module contains the package information for Inspyre-Toolbox.

Constants:
    GITHUB_URL (str):
        The URL to the Inspyre-Toolbox GitHub repository.

    PACKAGE_INFO (dict):
        The package information for Inspyre-Toolbox. This includes the name, included CLI entrypoints, URLs, and
        developer information. It has the following structure:

            {
                'name': {
                        'full':    'Inspyre-Toolbox',
                        'library': 'inspyre_toolbox',
                    },

                'included_cli_entrypoints': [
                        'ist-bytes-converter',
                        'ist-add-to-path'
                        ],

                'version': {
                            'semantic': str(VERSION),
                            'full': VERSION.to_full_version_string(),
                            'dict': {
                                        'major':       VERSION.major,
                                        'minor':       VERSION.minor,
                                        'patch':       VERSION.patch,
                                        'release':     VERSION.release,
                                        'release_num': VERSION.release_num
                                    },
                            'on_pypi': PYPI_VERSION_INFO,
                        },

                'urls': {
                        'version_control': {
                                'github': {
                                        'main':     GHU,
                                        'releases': f"{GHU}/releases",
                                        'issues':   f"{GHU}/issues",
                                        'wiki':     f"{GHU}/wiki",
                                        },
                                },
                        'pypi': {
                                'main':     'https://pypi.org/project/inspyre-toolbox',
                                'releases': 'https://pypi.org/project/inspyre-toolbox/#history',
                                },
                        },

                'developer': {
                        'org':  SOFTWARE_ORG,
                        'lead': SOFTWARE_ORG['developers'][0]
                        }
            }

    PYPI_URL (str):
        The URL to the Inspyre-Toolbox PyPi page.

Since:
    v1.6.0
"""
from platformdirs import PlatformDirs

from inspyre_toolbox.common.about.author import SOFTWARE_ORG
from inspyre_toolbox.common.about.pypi import PYPI_VERSION_INFO
from inspyre_toolbox.common.about.version import VERSION

PACKAGE_NAME = 'Inspyre-Toolbox'
GITHUB_URL = f"{SOFTWARE_ORG['urls']['github']}/{PACKAGE_NAME}"
PYPI_URL =   SOFTWARE_ORG['urls']['pypi']

PLATFORM_DIRS = PlatformDirs(appname=PACKAGE_NAME, appauthor=SOFTWARE_ORG['name'])

GHU = GITHUB_URL

PACKAGE_INFO = {
    'name': {
        'full': 'Inspyre-Toolbox',
        'library': 'inspyre_toolbox',
    },

    'version': {
        'semantic': str(VERSION),
        'full': VERSION.to_full_version_string(),
        'dict': {
            'major': VERSION.major,
            'minor': VERSION.minor,
            'patch': VERSION.patch,
            'release': VERSION.release,
            'release_num': VERSION.release_num
        },
        'on_pypi': PYPI_VERSION_INFO,
    },

    'included_cli_entrypoints': [
        'ist-bytes-converter',
        'ist-add-to-path'
    ],

    'urls': {
        'version_control': {
            'github': {
                'main': GHU,
                'releases': f"{GHU}/releases",
                'issues': f"{GHU}/issues",
                'wiki': f"{GHU}/wiki",
            },
        },
        'pypi': {
            'main': 'https://pypi.org/project/inspyre-toolbox',
            'releases': 'https://pypi.org/project/inspyre-toolbox/#history',
        },
    },

    'developer': {
        'org': SOFTWARE_ORG,
        'lead': SOFTWARE_ORG['developers']['lead']
    }
}

# Cleanup
del GHU
del SOFTWARE_ORG


__all__ = [
    'GITHUB_URL',
    'PACKAGE_INFO',
    'PLATFORM_DIRS',
    'PYPI_URL'
]



"""
This module provides shortcuts to the different option classes.

Classes:
    Option:
        The base class for all options.

    OptionLedger:
        A class to represent an option ledger.

    OptionLedgerEntry:
        A class to represent an entry in an option ledger

    OptionManager:
        A class to manage various options, acting like a dictionary with type enforcement

    YesNoOption:
        A class to represent a yes/no option.

    StringInputOption:
        A class to represent a string input option with a description.

Since:
    1.6.0
"""

from inspyre_toolbox.config_man.options import Option
from inspyre_toolbox.config_man.options.classes.string_input import StringInputOption
from inspyre_toolbox.config_man.options.classes.yes_no import YesNoOption
from inspyre_toolbox.config_man.options.ledger import OptionLedger, OptionLedgerEntry
from inspyre_toolbox.config_man.options.manager import OptionManager

__all__ = [
        'Option',
        'OptionLedger',
        'OptionLedgerEntry',
        'OptionManager',
        'YesNoOption',
        'StringInputOption'
        ]



from inspyre_toolbox.common.errors import InspyreToolboxError


class PyPiError(InspyreToolboxError):
    """
    Base class for PyPi errors.
    """
    default_message = 'An unknown error has occurred while interacting with PyPi.'

    def __init__(self, **kwargs):
        msg = kwargs.pop('message', self.default_message)
        super().__init__(message=msg, **kwargs)


class PyPiPackageNotFoundError(PyPiError):
    """
    Raised when a package is not found on PyPi.
    """
    default_message = 'The package was not found on PyPi.'

    def __init__(self, **kwargs):
        message = kwargs.pop('message', self.default_message)
        super().__init__(message=message, **kwargs)



import os
from warnings import warn

from inspyre_toolbox.path_man import check_file, prepare_path
from inspyre_toolbox.sys_man.operating_system import MOD_LOGGER as PARENT_LOGGER
from inspyre_toolbox.sys_man.operating_system.linux.system import run_as_admin
from inspyre_toolbox.sys_man.operating_system.linux.users import is_admin

MOD_LOGGER = PARENT_LOGGER.get_child('linux')

IS_ADMIN = is_admin()

del is_admin

PROFILE_FILE = os.getenv('HOME') + '/.bashrc'


def add_to_path(new_path: str, do_not_notify=None, profile_file: str = PROFILE_FILE):
    """
    The add_to_path function adds the program path to the system's PATH variable on Unix-like systems.
    This allows for easy access to programs from any terminal window.

    Args:
        new_path (str):
            The full path to the directory you'd like to add to your system's PATH
            environment variable.

        do_not_notify (bool):
            This parameter is not used on Unix-like systems. Ignoring parameter.

        profile_file (str):
            The file to add the path to. Defaults to ~/.bashrc.

    Returns:
        None
    """
    if do_not_notify:
        warn("The do_not_notify parameter is not used on Unix-like systems. Ignoring parameter.")

    new_path = str(prepare_path(new_path))
    profile_file = str(prepare_path(profile_file))

    if not check_file(profile_file):
        raise FileNotFoundError(f"File {profile_file} does not exist")

    with open(profile_file, "a") as bash_file:  # Open bashrc file
        bash_file.write(f'\nexport PATH="{new_path}:$PATH"\n')  # Ad
        # d program path to Path variable
    os.system(f". {os.getenv('HOME')}/.bashrc")  # Update bash source
    print(f"Added {new_path} to path on Unix-like system, please restart shell for changes to take effect")
    print('"exec $SHELL" to restart shell')


__all__ = [
        'IS_ADMIN',
    'add_to_path',
    'run_as_admin'
        ]



class FreezeProperty:
    """
    A descriptor that provides a property which can only be set once.

    Attributes:
        fset (function): The setter function for the property.
        is_set (bool): A flag to indicate if the property has been set.
    """

    def __init__(self, fset):
        """
        Initialize the FreezeProperty.

        Args:
            fset (function): The setter function for the property.
        """
        self.fset = fset
        self.is_set = False

    def __call__(self, instance, value):
        """
        Set the value of the property. Can only be done once.

        Args:
            instance (object): The instance to which the property belongs.
            value: The value to set for the property.

        Raises:
            AttributeError: If the property has already been set.
        """
        if self.is_set:
            raise AttributeError(
                f"{instance.__class__.__name__}.{self.fset.__name__} is frozen and cannot be modified.")
        self.fset(instance, value)
        self.is_set = True

    def __get__(self, instance, owner):
        """
        Get the value of the property.

        Args:
            instance (object): The instance to which the property belongs.
            owner (type): The class to which the property belongs.

        Returns:
            The value of the property.
        """
        if instance is None:
            return self
        return instance.__dict__.get(self.fset.__name__, None)

    def __set__(self, instance, value):
        """
        Set the value of the property. Can only be done once.

        Args:
            instance (object): The instance to which the property belongs.
            value: The value to set for the property.

        Raises:
            AttributeError: If the property has already been set.
        """
        if self.is_set:
            raise AttributeError(f"Property {self.fset.__name__} is frozen and cannot be modified.")
        self.__call__(instance, value)


def freeze_property(fset):
    """
    Create a FreezeProperty instance.

    Args:
        fset (function): The setter function for the property.

    Returns:
        FreezeProperty: A new FreezeProperty instance.
    """
    return FreezeProperty(fset)



import inspect

from .errors import UnmetConditionError


class RestrictedSetter:
    """
    A descriptor class that restricts the setting of a property based on type, value, and custom conditions.

    This class is designed to be used as a descriptor for class classes, allowing for the definition of restrictions
    on the setting of the property. These restrictions can include type checking, value checking, and custom conditions
    that must be met for the property to be set. If any of the restrictions are violated, an exception is raised.

    The `RestrictedSetter` class also allows for the definition of an initial value for the property, which is used if
    the property has not been set. This can be useful for defining default values for classes.

    Notes:
      - If a `preferred_type` is specified, the value will be converted to that type if possible. If the value cannot be converted to the preferred type, a `TypeError` will be raised.

      - If a custom condition is specified, the condition must be met for the property to be set. If the condition is not met, an `UnmetConditionError` will be raised.

      - If an exception is specified, it will be raised instead of the default `UnmetConditionError` if the custom condition is not met.

      - If `allowed_values` is specified, the value must be in the set of allowed values for the property to be set. If the value is not in the set, a `ValueError` will be raised.

      - If `restrict_setter` is set to `True`, the setter can only be called within class methods. If the setter is called outside of class methods, a `PermissionError` will be raised.
    """

    def __init__(
            self,
            name,
            initial=None,
            allowed_types=None,
            preferred_type=None,
            allowed_values=None,
            condition=None,
            exception=None,
            exception_args=None,
            restrict_setter=True
    ):
        """
        Initialize a new RestrictedSetter instance.

        Parameters:

            name (str):
                The name of the property.

            initial (Any, optional):
                The initial value of the property.

            allowed_types (Union[type, Tuple[type]], optional):
                The type(s) allowed for the property.

            preferred_type (type, optional):
                The preferred type for the property.

            allowed_values (Iterable, optional):
                The set of allowed values for the property.

            condition (Callable, optional):
                A condition that must be met for the property to be set.

            exception (Exception, optional):
                The exception to raise if the condition is not met.

            exception_args (Dict, optional):
                Arguments to pass to the exception.

            restrict_setter (bool, optional):
                Whether to restrict the setter to class methods.


        Example:

                >>> class MyClass:
                ...     _my_property = RestrictedSetter(
                ...         'my_property',
                ...         allowed_types=(int, float),
                ...         preferred_type=int,
                ...         condition=lambda self: self._my_property > 0,
                ...         exception=ValueError,
                ...         exception_args={'message': "Value must be greater than 0."}
                ...     )
                ...
                ...     @property
                ...     def my_property(self):
                ...         return self._my_property
                ...
                ...     @my_property.setter
                ...     def my_property(self, value):
                ...         self._my_property = value
                ...
                >>> obj = MyClass()
                >>> obj.my_property = 10

        Example:
            >>> class MyClass:
            ...     _my_property = RestrictedSetter(
            ...         'my_property',
            ...         allowed_types=(int, float),
            ...         preferred_type=int,
            ...         allowed_values={10, 20, 30},
            ...         condition=lambda self: self._my_property > 0,
            ...         exception=ValueError,
            ...         exception_args={'message': "Value must be greater than 0."}
            ...     )
            ...
            ...     @property
            ...     def my_property(self):
            ...         return self._my_property
            ...
            ...     @my_property.setter
            ...     def my_property(self, value):
            ...         self._my_property = value
            ...
            >>> obj = MyClass()
            >>> obj.my_property = 10  # Valid
            >>> obj.my_property = 5   # Raises an exception if 5 is not in allowed_values
        """

        self.name = f'_{name}'
        self.initial = initial
        self.restrict_setter = restrict_setter

        # Ensure `allowed_types` is a tuple or list
        if not isinstance(allowed_types, (tuple, list)):
            allowed_types = (allowed_types,)

        self.allowed_types = allowed_types if isinstance(allowed_types, (tuple, list)) else (allowed_types,)
        self.preferred_type = preferred_type

        self.allowed_values = (
            allowed_values if isinstance(
                allowed_values, (tuple, list)) else (allowed_values,)
        ) if allowed_values is not None else None

        self.condition = condition
        self.exception = exception
        self.exception_args = exception_args or {}

    def __get__(self, obj, obj_type=None):
        """
        Retrieves the value of the property.

        This method is a getter for the property, returning its current value. If the property has not been set,
        it returns the initial value specified during the creation of the `RestrictedSetter` instance.

        Parameters:

            obj (object):
                The instance of the class whose property is being retrieved.

            obj_type (type, optional):
                The type of the object. This parameter is not used but is part of the descriptor protocol.

        Returns:
            The current value of the property.
        """
        return getattr(obj, self.name, self.initial)

    def __set__(self, obj, value):
        """
        Sets the value of the property, enforcing any restrictions.

        This method acts as a setter for the property, applying the defined restrictions such as type checking,
        value restrictions, and custom conditions. If any restriction is violated, an appropriate exception is raised.

        Parameters:

            obj (object):
                The instance of the class whose property is being set.

            value (Any):
                The new value for the property.

        Raises:
            PermissionError:
                If the setter is called outside-of class methods when `restrict_setter` is True.

            ValueError:
                If the provided value is not in the `allowed_values` set.

            PermissionError:
                If a custom `condition` is not satisfied.

            TypeError:
                If the provided value is not of an allowed type.

            Any exception specified by the `exception` attribute,
                if the custom condition fails.
        """
        if (self.allowed_types and not isinstance(value, self.allowed_types)) and value is not None:
            raise TypeError(
                f"Value must be of type {', '.join([t.__name__ for t in self.allowed_types])}, got type {type(value).__name__}.")

        if self.preferred_type and not isinstance(value, self.preferred_type):
            value = self.preferred_type(value)

        if self.restrict_setter:
            caller_frame = inspect.stack()[1]
            caller_self = caller_frame.frame.f_locals.get('self', None)
            if caller_self is None or caller_self.__class__ != obj.__class__:
                raise PermissionError("Property can only be set within class methods.")

        if self.allowed_values is not None and value not in self.allowed_values:
            allowed_vals = ', '.join(map(str, self.allowed_values))
            raise ValueError(f"Value must be one of {allowed_vals}, got {value}")

        if self.condition and not self.condition(obj):
            if self.exception:
                raise self.exception(**self.exception_args)
            raise UnmetConditionError("Condition for setting property not met.")

        setattr(obj, self.name, value)

    def __delete__(self, obj):
        """
        Resets the property to its initial value.

        This method allows for the property to be reset or deleted, setting it back to the initial value specified
        during the creation of the `RestrictedSetter` instance. This can be useful for resetting the state of an
        instance property to its default.

        Parameters:

            obj (object):
                The instance of the class whose property is being reset.
        """
        setattr(obj, self.name, self.initial)
